"""
Module for Cloudformation-related resources.
"""
import datetime

from .arn import ARN
from .base_control import (
    Describer,
    MultiLister,
    NoResults,
    ResourceLister,
    format_timedelta,
    tag_finder_generator,
)
from .common import Common
from .resource_common import multilister_with_compare_path


class CFNDescriber(Describer):
    """
    Describer control for Cloudformation Stack resources.
    """

    prefix = "cfn_browser"
    title = "CloudFormation Stack"

    resource_type = "stack"
    main_provider = "cloudformation"
    category = "Cloudformation"
    subcategory = "Stack"
    describe_method = "describe_stacks"
    describe_kwarg_name = "StackName"
    object_path = ".Stacks[0]"

    def populate_entry(self, *args, **kwargs):
        super().populate_entry(*args, **kwargs)
        try:
            arn = ARN(self.entry_id)
            self.entry_id = arn.resource_id_first
        except ValueError:
            return


# TODO: Refactor
class CFNRelated(MultiLister):
    """
    Related resource lister for Cloudformation Stack resources.
    """

    prefix = "cfn_related"
    title = "Resources in CloudFormation Stack"

    def title_info(self):
        return self.compare_value

    def __init__(self, *args, **kwargs):
        self.stack_res_list = {}
        kwargs["compare_key"] = "arn"
        self.resource_descriptors = [
            multilister_with_compare_path(
                "ec2", '.Tags[] | select(.Key=="aws:cloudformation:stack-id").Value'
            ),
            multilister_with_compare_path(
                "rds", '.TagList[] | select(.Key=="aws:cloudformation:stack-id").Value'
            ),
            {
                "resource_key": "autoscaling",
                "list_method": "describe_auto_scaling_groups",
                "list_kwargs": {},
                "item_path": ".AutoScalingGroups",
                "column_paths": {
                    "type": lambda x: "Autoscaling Group",
                    "id": self.empty,
                    "name": ".AutoScalingGroupName",
                },
                "hidden_columns": {},
                "compare_as_list": False,
                "compare_path": '.Tags[] | select(.Key=="aws:cloudformation:stack-id").Value',
            },
            {
                "resource_key": "elbv2",
                "list_method": "describe_load_balancers",
                "list_kwargs": {},
                "item_path": ".LoadBalancers",
                "column_paths": {
                    "type": lambda x: "Load Balancer",
                    "id": self.empty,
                    "name": ".LoadBalancerName",
                },
                "hidden_columns": {
                    "arn": ".LoadBalancerArn",
                },
                "compare_as_list": False,
                "compare_path": self.comparer_generator(
                    "AWS::ElasticLoadBalancingV2::LoadBalancer", ".LoadBalancerArn"
                ),
            },
            {
                "resource_key": "elbv2",
                "list_method": "describe_target_groups",
                "list_kwargs": {},
                "item_path": ".TargetGroups",
                "column_paths": {
                    "type": lambda x: "Target Group",
                    "id": self.resource_id_from_arn_generator(".TargetGroupArn"),
                    "name": ".TargetGroupName",
                },
                "hidden_columns": {
                    "arn": ".TargetGroupArn",
                },
                "compare_as_list": False,
                "compare_path": self.comparer_generator(
                    "AWS::ElasticLoadBalancingV2::TargetGroup", ".TargetGroupArn"
                ),
            },
            {
                "resource_key": "elbv2",
                "list_method": "describe_listeners",
                "list_kwargs": self.kwargs_from_physids_generator(
                    "ListenerArns", "AWS::ElasticLoadBalancingV2::Listener"
                ),
                "item_path": ".Listeners",
                "column_paths": {
                    "type": lambda x: "Listener",
                    "id": self.full_resource_id_from_arn_generator(".ListenerArn"),
                    "name": self.empty,
                },
                "hidden_columns": {
                    "arn": ".ListenerArn",
                },
                "compare_as_list": False,
                "compare_path": self.comparer_generator(
                    "AWS::ElasticLoadBalancingV2::Listener", ".ListenerArn"
                ),
            },
            {
                "resource_key": "ec2",
                "list_method": "describe_vpcs",
                "list_kwargs": {},
                "item_path": ".Vpcs",
                "column_paths": {
                    "type": lambda x: "VPC",
                    "id": ".VpcId",
                    "name": tag_finder_generator("Name"),
                },
                "hidden_columns": {},
                "compare_as_list": False,
                "compare_path": '.Tags[] | select(.Key=="aws:cloudformation:stack-id").Value',
            },
            {
                "resource_key": "ec2",
                "list_method": "describe_subnets",
                "list_kwargs": {},
                "item_path": ".Subnets",
                "column_paths": {
                    "type": lambda x: "VPC Subnet",
                    "id": ".SubnetId",
                    "name": tag_finder_generator("Name"),
                },
                "hidden_columns": {},
                "compare_as_list": False,
                "compare_path": '.Tags[] | select(.Key=="aws:cloudformation:stack-id").Value',
            },
            {
                "resource_key": "route53",
                "list_method": "list_hosted_zones",
                "list_kwargs": {},
                "item_path": ".HostedZones",
                "column_paths": {
                    "type": lambda x: "Route53 Zone",
                    "id": ".Id",
                    "name": ".Name",
                },
                "hidden_columns": {},
                "compare_as_list": False,
                "compare_path": self.comparer_generator(
                    "AWS::Route53::HostedZone", ".Id"
                ),
            },
            {
                "resource_key": "ec2",
                "list_method": "describe_vpcs",
                "list_kwargs": {},
                "item_path": ".Vpcs",
                "column_paths": {
                    "type": lambda x: "VPC",
                    "id": ".VpcId",
                    "name": tag_finder_generator("Name"),
                },
                "hidden_columns": {},
                "compare_as_list": False,
                "compare_path": '.Tags[] | select(.Key=="aws:cloudformation:stack-id").Value',
            },
            {
                "resource_key": "ec2",
                "list_method": "describe_subnets",
                "list_kwargs": {},
                "item_path": ".Subnets",
                "column_paths": {
                    "type": lambda x: "Subnet",
                    "id": ".SubnetId",
                    "name": tag_finder_generator("Name"),
                },
                "hidden_columns": {},
                "compare_as_list": False,
                "compare_path": '.Tags[] | select(.Key=="aws:cloudformation:stack-id").Value',
            },
            {
                "resource_key": "ec2",
                "list_method": "describe_route_tables",
                "list_kwargs": {},
                "item_path": ".RouteTables",
                "column_paths": {
                    "type": lambda x: "Route Table",
                    "id": ".RouteTableId",
                    "name": tag_finder_generator("Name"),
                },
                "hidden_columns": {},
                "compare_as_list": False,
                "compare_path": '.Tags[] | select(.Key=="aws:cloudformation:stack-id").Value',
            },
        ]
        super().__init__(*args, **kwargs)

    def full_resource_id_from_arn_generator(self, arn_path):
        """
        Column callback factory. Returns a function which retrieves the full resource ID from an ARN.

        Parameters
        ----------
        arn_path : str
            The jq path for the ARN.

        Returns
        -------
        callable(dict) -> str
            A function which retrieves the full resource ID from the ARN at arn_path.
        """

        def fn(raw_item):
            arn = ARN(Common.Session.jq(arn_path).input(raw_item).first())
            return arn.resource_id

        return fn

    def resource_id_from_arn_generator(self, arn_path):
        """
        Column callback factory. Returns a function which retrieves the first resource ID from an ARN.

        Parameters
        ----------
        arn_path : str
            The jq path for the ARN.

        Returns
        -------
        callable(dict) -> str
            A function which retrieves the first resource ID from the ARN at arn_path.
        """

        def fn(raw_item):
            arn = ARN(Common.Session.jq(arn_path).input(raw_item).first())
            return arn.resource_id_first

        return fn

    def comparer_generator(self, cfn_type, physical_id_path):
        """
        Comparer function generator. The comparer function generated by this generator determines whether a resource's physical ID is present
        in the list of physical resource IDs associated with the stack.

        If a MultiLister descriptor's compare_path is a function, the return value of that function is used for the comparison rather than the
        result of jq applied to the input. This enables us to use custom, more complex comparisons where required, while also retaining the
        ability to use very generic functions (such as one that simply returns the ID of the cloudformation stack a resource is associated
        with). For this reason, this comparer_generator can generate a compare_path function for us.

        Parameters
        ----------
        cfn_type : str
            Cloudformation type specification for the resources being compared by the comparer.
        physical_id_path : str
            The jq path where the resource's physical ID is available.

        Returns
        -------
        callable(dict) -> str
            A function which returns the stack ID of this stack if the resource is present in the stack, or None otherwise.
        """

        def fn(raw_item):
            phys_id = Common.Session.jq(physical_id_path).input(raw_item).first()
            if (
                cfn_type in self.stack_res_list
                and phys_id in self.stack_res_list[cfn_type]
            ):
                return self.compare_value
            return None

        return fn

    def kwargs_from_physids_generator(self, kwarg, cfn_type):
        """
        Kwargs callback factory. Returns a function which returns a list of physical IDs for stack resources with a certain Cloudformation
        type, as a dict, with a specified key.

        Parameters
        ----------
        kwarg : str
            The key for the resource ID list in the returned dict.
        cfn_type : str
            The Cloudformation type for which resource IDs will be listed.

        Returns
        -------
        callable() -> dict
            A function which returns a map mapping kwarg to the list of physical IDs of resources of type cfn_type.
        """

        def fn():
            if cfn_type in self.stack_res_list:
                return {kwarg: self.stack_res_list[cfn_type]}
            raise NoResults

        return fn

    def async_inner(self, *args, fn, clear=False, **kwargs):
        resource_list = {}
        self.stack_res_list = {}
        while True:
            rargs = {"StackName": self.orig_compare_value["name"]}
            if "NextToken" in resource_list and resource_list["NextToken"] is not None:
                rargs["NextToken"] = resource_list["NextToken"]
            resource_list = Common.Session.service_provider(
                "cloudformation"
            ).list_stack_resources(**rargs)
            for item in resource_list["StackResourceSummaries"]:
                if item["ResourceType"] not in self.stack_res_list:
                    self.stack_res_list[item["ResourceType"]] = []
                self.stack_res_list[item["ResourceType"]].append(
                    item["PhysicalResourceId"]
                )
            if "NextToken" not in resource_list or resource_list["NextToken"] is None:
                break
        return super().async_inner(*args, fn=fn, clear=clear, **kwargs)


def _cfn_determine_created(cfn):
    """
    Column callback for extracting when a stack was created.
    """
    return format_timedelta(
        datetime.datetime.now(datetime.timezone.utc)
        - datetime.datetime.fromisoformat(cfn["CreationTime"])
    )


def _cfn_determine_updated(cfn):
    """
    Column callback for extracting when a stack was updated.
    """
    if "LastUpdatedTime" in cfn:
        return format_timedelta(
            datetime.datetime.now(datetime.timezone.utc)
            - datetime.datetime.fromisoformat(cfn["LastUpdatedTime"])
        )
    return _cfn_determine_created(cfn)


class CFNResourceLister(ResourceLister):
    """
    Lister control for Cloudformation Stack resources.
    """

    prefix = "cfn_list"
    title = "CloudFormation Stacks"
    command_palette = ["cfn", "cloudformation"]

    resource_type = "stack"
    main_provider = "cloudformation"
    category = "Cloudformation"
    subcategory = "Stack"
    list_method = "describe_stacks"
    item_path = ".Stacks"
    columns = {
        "name": {"path": ".StackName", "size": 30, "weight": 0, "sort_weight": 0},
        "status": {"path": ".StackStatus", "size": 15, "weight": 1},
        "drift": {
            "path": ".DriftInformation.StackDriftStatus",
            "size": 15,
            "weight": 2,
        },
        "created": {"path": _cfn_determine_created, "size": 20, "weight": 3},
        "updated": {"path": _cfn_determine_updated, "size": 20, "weight": 4},
        "arn": {"path": ".StackId", "hidden": True},
    }
    describe_command = CFNDescriber.opener
    open_command = CFNRelated.opener
    open_selection_arg = "compare_value"


def _cfn_determine_parameter_value(param):
    if param["ParameterValue"] == "*" * len(param["ParameterValue"]):
        return "<hidden>"
    return param["ParameterValue"]


@ResourceLister.Autocommand("CFNResourceLister", "p", "Parameters", "stack")
class CFNParameters(ResourceLister):
    """
    Lister control for Cloudformation Stack parameters.
    """

    prefix = "cfn_list"
    title = "CloudFormation Stack Parameters"

    resource_type = "stack parameter"
    main_provider = "cloudformation"
    category = "Cloudformation"
    subcategory = "Stack Parameters"
    list_method = "describe_stacks"
    item_path = ".Stacks[0].Parameters"
    columns = {
        "name": {"path": ".ParameterKey", "size": 30, "weight": 0, "sort_weight": 0},
        "value": {"path": _cfn_determine_parameter_value, "size": 30, "weight": 1},
    }

    def title_info(self):
        return self.stack_name

    def __init__(self, *args, stack=None, **kwargs):
        self.stack_name = stack["name"]
        self.list_kwargs = {"StackName": stack["name"]}
        super().__init__(*args, **kwargs)


@ResourceLister.Autocommand("CFNResourceLister", "o", "Outputs", "stack")
class CFNOutputs(ResourceLister):
    """
    Lister control for Cloudformation Stack outputs.
    """

    prefix = "cfn_list"
    title = "CloudFormation Stack Outputs"

    resource_type = "stack output"
    main_provider = "cloudformation"
    category = "Cloudformation"
    subcategory = "Stack Outputs"
    list_method = "describe_stacks"
    item_path = ".Stacks[0].Outputs"
    columns = {
        "name": {"path": ".OutputKey", "size": 30, "weight": 0, "sort_weight": 0},
        "value": {"path": ".OutputValue", "size": 30, "weight": 1},
    }

    def title_info(self):
        return self.stack_name

    def __init__(self, *args, stack=None, **kwargs):
        self.stack_name = stack["name"]
        self.list_kwargs = {"StackName": stack["name"]}
        super().__init__(*args, **kwargs)



"""
AWS Autoscaling Group resource controls.
"""
from .base_control import Describer, ResourceLister, SelectionAttribute, TemplateDict
from .common import Common, SessionAwareDialog
from .termui.control import Border
from .termui.dialog import DialogFieldCheckbox, DialogFieldLabel, DialogFieldText
from .termui.ui import ControlCodes


def _asg_determine_launch_info(asg):
    """
    Column callback for determining the name of the launch configuration or launch template associated with an autoscaling group.

    Parameters
    ----------
    asg : dict
        Partial raw JSON response from the AWS API, one autoscaling group object.

    Returns
    -------
    str
        The name of the launch configuration or launch template associated with the ASG, or empty if not found.
    """
    if "LaunchConfigurationName" in asg and bool(asg["LaunchConfigurationName"]):
        return asg["LaunchConfigurationName"]
    if "LaunchTemplate" in asg:
        return asg["LaunchTemplate"]["LaunchTemplateName"]
    return ""


def _asg_determine_instance_count(asg):
    """
    Column callback for determining the number of healthy and total instances in an autoscaling group.

    Parameters
    ----------
    asg : dict
        Partial raw JSON response from the AWS API, one autoscaling group object.

    Returns
    -------
    str
        The number of healthy and total instances in the group in the format of healthy/total.
    """
    healthy = len([h for h in asg["Instances"] if h["HealthStatus"] == "Healthy"])
    total = len(asg["Instances"])
    return f"{healthy}/{total}"


class ASGDescriber(Describer):
    """
    Describer control for Autoscaling Group resources.
    """

    prefix = "asg_browser"
    title = "Autoscaling Group"

    resource_type = "autoscaling group"
    main_provider = "autoscaling"
    category = "Autoscaling"
    subcategory = "Autoscaling Group"
    describe_method = "describe_auto_scaling_groups"
    describe_kwarg_name = "AutoScalingGroupNames"
    describe_kwarg_is_list = True
    object_path = ".AutoScalingGroups[0]"


class ASGResourceLister(ResourceLister):
    """
    Lister control for Autoscaling Group resources.

    Attributes
    ----------
    launch_config : awsc.termui.list_control.ListEntry
        List ASGs in the context of this launch configuration.
    """

    prefix = "asg_list"
    title = "Autoscaling Groups"
    command_palette = ["asg", "autoscaling"]

    resource_type = "autoscaling group"
    main_provider = "autoscaling"
    category = "Autoscaling"
    subcategory = "Autoscaling Group"
    list_method = "describe_auto_scaling_groups"
    item_path = ".AutoScalingGroups"
    columns = {
        "name": {
            "path": ".AutoScalingGroupName",
            "size": 30,
            "weight": 0,
            "sort_weight": 0,
        },
        "template": {
            "path": _asg_determine_launch_info,
            "size": 30,
            "weight": 1,
        },
        "current": {
            "path": _asg_determine_instance_count,
            "size": 10,
            "weight": 2,
        },
        "min": {"path": ".MinSize", "size": 10, "weight": 3},
        "desired": {"path": ".DesiredCapacity", "size": 10, "weight": 4},
        "max": {"path": ".MaxSize", "size": 10, "weight": 5},
        "launch config": {"path": ".LaunchConfigurationName", "hidden": True},
        "arn": {"path": ".AutoScalingGroupARN", "hidden": True},
    }

    describe_command = ASGDescriber.opener
    open_command = "i"
    primary_key = "name"

    def title_info(self):
        return (
            None
            if self.launch_config is None
            else f"LaunchConfiguration: {self.launch_config['name']}"
        )

    def matches(self, list_entry, *args):
        if self.launch_config is not None:
            if list_entry["launch config"] != self.launch_config["name"]:
                return False
        return super().matches(list_entry, *args)

    def __init__(self, *args, lc=None, **kwargs):
        self.launch_config = lc
        super().__init__(*args, **kwargs)

    @ResourceLister.Autohotkey(ControlCodes.S, tooltip="Scale", is_validated=True)
    def scale_group(self, _):
        """
        Hotkey callback for resize action on autoscaling group.
        """
        ASGScaleDialog.opener(
            caller=self,
        )

    @ResourceLister.Autohotkey(ControlCodes.T, tooltip="Rollout", is_validated=True)
    def rollout(self, _):
        self.confirm_template(
            "start_instance_refresh",
            TemplateDict(
                {
                    "AutoScalingGroupName": SelectionAttribute("name"),
                }
            ),
            action_name="Rollout",
        )(self.selection)

    @ResourceLister.Autohotkey(
        ControlCodes.Z, tooltip="Cancel Rollout", is_validated=True
    )
    def cancel_rollout(self, _):
        self.confirm_template(
            "cancel_instance_refresh",
            TemplateDict(
                {
                    "AutoScalingGroupName": SelectionAttribute("name"),
                }
            ),
            action_name="Cancel rollout for",
        )(self.selection)


class ASGScaleDialog(SessionAwareDialog):
    """
    Dialog for scaling autoscaling groups.

    Attributes
    ----------
    asg_entry : awsc.termui.list_control.ListEntry
        The autoscaling group being scaled.
    error_label : awsc.termui.dialog.DialogFieldLabel
        An error display field, for displaying validation errors.
    desired_capacity_field : awsc.termui.dialog.DialogFieldText
        Textfield for inputting the new desired capacity of the group.
    adjust_limits_field : awsc.termui.dialog.DialogFieldCheckbox
        If checked, desired capacity may exceed bounds set by min/max size, as those will be automatically expanded to fit the new desired
        capacity.
    caller : awsc.termui.control.Control
        The parent control opening this dialog.
    """

    def __init__(self, *args, caller=None, **kwargs):
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            "Scale autoscaling group",
            Common.color("modal_dialog_border_title"),
            caller.selection["name"],
            Common.color("modal_dialog_border_title_info"),
        )
        self.asg_entry = caller.selection
        super().__init__(caller=caller, *args, **kwargs)
        self.desired_capacity_field = DialogFieldText(
            "Desired capacity:",
            text=str(self.asg_entry["desired"]),
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
            accepted_inputs="0123456789",
        )
        self.add_field(self.desired_capacity_field)
        self.add_field(DialogFieldLabel(""))
        self.adjust_limits_field = DialogFieldCheckbox(
            "Adjust min/max capacity if required",
            checked=True,
            color=Common.color("modal_dialog_textfield_label"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
        )
        self.add_field(self.adjust_limits_field)
        self.caller = caller

    def accept_and_close(self):
        if self.desired_capacity_field.text == "":
            self.error_label.text = "Desired capacity cannot be blank."
            return

        b3s = Common.Session.service_provider("autoscaling")
        asg = b3s.describe_auto_scaling_groups(
            AutoScalingGroupNames=[self.asg_entry["name"]]
        )["AutoScalingGroups"][0]
        des = int(self.desired_capacity_field.text)

        if (
            des < asg["MinSize"] or des > asg["MaxSize"]
        ) and not self.adjust_limits_field.checked:
            self.error_label.text = f"Desired capacity is out of min-max range of {asg['MinSize']}-{asg['MaxSize']}"
            return

        nmin = min(des, asg["MinSize"])
        nmax = max(des, asg["MaxSize"])

        b3s.update_auto_scaling_group(
            AutoScalingGroupName=self.asg_entry["name"],
            DesiredCapacity=des,
            MinSize=nmin,
            MaxSize=nmax,
        )
        super().accept_and_close()

    def close(self):
        if self.caller is not None:
            self.caller.refresh_data()
        super().close()



"""
Main module for the awsc application.
"""
import argparse
import configparser
import os
import sys
from pathlib import Path

import boto3

# pylint: disable=unused-import # We use this import to enumerate all resource listers.
from . import resources
from .aws import AWS
from .commander import Commander, Filterer
from .common import Common, DefaultAnchor, DefaultDimension
from .context import ContextList
from .dashboard import Dashboard
from .log import LogLister
from .meta import CommanderOptionsLister
from .region import RegionList
from .ssh import SSHList
from .termui.alignment import Dimension, TopLeftAnchor
from .termui.control import Border


def awscheck():
    """
    Predicate function for checking if AWS is initialized.
    """
    return bool(Common.Session.context) and bool(Common.Session.region)


def open_filterer():
    """
    Hotkey callback for opening the filter bar.
    """
    if Common.Session.filterer is None:
        return Filterer(
            Common.Session.ui.top_block,
            TopLeftAnchor(0, 8),
            Dimension("100%", "3"),
            session=Common.Session,
            color=Common.color("search_bar_color"),
            symbol_color=Common.color("search_bar_symbol_color"),
            inactive_color=Common.color("search_bar_inactive_color"),
            weight=-200,
            border=Border(
                Common.border("search_bar"), Common.color("search_bar_border")
            ),
        )
    return Common.Session.filterer.resume()


def open_commander():
    """
    Hotkey callback for opening the command palette.
    """
    return Commander(
        Common.Session.ui.top_block,
        TopLeftAnchor(0, 8),
        Dimension("100%", "3"),
        session=Common.Session,
        color=Common.color("command_bar_color"),
        symbol_color=Common.color("command_bar_symbol_color"),
        autocomplete_color=Common.color("command_bar_autocomplete_color"),
        ok_color=Common.color("command_bar_ok_color"),
        error_color=Common.color("command_bar_error_color"),
        weight=-200,
        border=Border(Common.border("search_bar"), Common.color("search_bar_border")),
    )


def main(*args, **kwargs):
    """
    Entrypoint for awsc.
    """
    # stderr hack
    old_stderr = None
    try:
        Common.initialize()

        if os.fstat(0) == os.fstat(1):
            # pylint: disable=consider-using-with # With would be extremely roundabout here.
            log_file_handle = open(
                Common.Configuration["error_log"], "w", buffering=1, encoding="utf-8"
            )
            old_stderr = sys.stderr
            sys.stderr = log_file_handle
        Common.Session.service_provider = AWS()
        Common.post_initialize()
        if Common.Session.context == "":
            Common.Session.replace_frame(ContextList.opener())
        else:
            if Common.Session.context_data["mfa_device"] != "":
                Common.Session.replace_frame(ContextList.opener())
            else:
                Common.Session.replace_frame(Dashboard.opener())

        Common.Session.info_display.commander_hook = open_commander
        Common.Session.info_display.filterer_hook = open_filterer

        Common.Session.commander_options["ctx"] = ContextList.opener
        Common.Session.commander_options["context"] = ContextList.opener
        Common.Session.commander_options["region"] = RegionList.opener
        Common.Session.commander_options["ssh"] = SSHList.opener
        Common.Session.commander_options["logs"] = LogLister.opener
        Common.Session.commander_options["?"] = CommanderOptionsLister.opener
        Common.Session.commander_options["help"] = CommanderOptionsLister.opener

        Common.main()
    finally:
        if old_stderr is not None:
            sys.stderr.close()
            sys.stderr = old_stderr


def cred_helper(*args, **kwargs):
    parser = argparse.ArgumentParser(
        prog="AWSC Credentials Helper",
        description="Allows access to awsc credentials keystore for use with aws cli",
    )
    parser.add_argument("context")
    args = parser.parse_args()
    Common.initialize()
    conf = Common.Configuration
    if args.context not in conf.keystore:
        print(f"Keypair {args.context} not found.", file=sys.stderr)
    keypair = conf.keystore[args.context]
    sts = boto3.client(
        "sts",
        aws_access_key_id=keypair["access"],
        aws_secret_access_key=keypair["secret"],
    )
    context = conf["contexts"][args.context]
    if context["mfa_device"] != "":
        mfa_code = input(f"Enter MFA code for device {context['mfa_device']}: ")
        resp = sts.get_session_token(
            DurationSeconds=86400,
            SerialNumber=context["mfa_device"],
            TokenCode=mfa_code,
        )
    else:
        resp = sts.get_session_token(DurationSeconds=86400)

    aws_creds = Path.home() / ".aws" / "credentials"
    try:
        with aws_creds.open("r", encoding="utf-8") as file:
            creds = file.read()
    except OSError as error:
        if error.errno == 2:  # FNF
            creds = ""
        else:
            print(
                f"Failed to open ~/.aws/credentials: {str(error)}",
                file=sys.stderr,
            )
            return

    parser = configparser.ConfigParser(default_section="__default")
    parser.read_string(creds)
    if parser.has_section(args.context):
        parser.remove_section(args.context)
    parser.add_section(args.context)

    token = resp["Credentials"]["SessionToken"]
    expiration = resp["Credentials"]["Expiration"]
    parser.set(args.context, "aws_session_token", token)
    parser.set(args.context, "aws_security_token", token)
    parser.set(
        args.context,
        "expiration",
        expiration.strftime("%Y-%m-%d %H:%M:%S"),
    )
    parser.set(
        args.context,
        "aws_access_key_id",
        resp["Credentials"]["AccessKeyId"],
    )
    parser.set(
        args.context,
        "aws_secret_access_key",
        resp["Credentials"]["SecretAccessKey"],
    )

    with aws_creds.open("w", encoding="utf-8") as file:
        parser.write(file)
    print(f"Wrote short term token to profile {args.context}", file=sys.stderr)



"""
Module for cloudwatch resources.
"""
import datetime

from .base_control import (
    Describer,
    ListResourceDocumentEditor,
    ResourceLister,
    SelectionAttribute,
)
from .common import BaseChart, Common
from .dashboard import KeyValueDashboardBlock
from .termui.ui import ControlCodes


class MetricViewer(BaseChart):
    """
    Metric display control. Displays the selected metric as a bar graph.
    """

    prefix = "metric_view"
    title = "Metrics"

    def title_info(self):
        return f"{self.metric_dimension['Value']} {self.metric_name}"

    def __init__(self, *args, caller, metric=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.metric_name = metric["name"]
        self.metric_namespace = metric["namespace"]
        self.metric_dimension = {
            "Name": caller.dimension[0],
            "Value": caller.dimension[1],
        }
        self.load_data()

    def load_data(self):
        """
        Loads time series into the bar graph based on the metric name, namespace and dimension provided.
        """
        api_kwargs = {
            "MetricDataQueries": [
                {
                    "Id": "metric",
                    "MetricStat": {
                        "Metric": {
                            "Namespace": self.metric_namespace,
                            "MetricName": self.metric_name,
                            "Dimensions": [self.metric_dimension],
                        },
                        "Period": 300,
                        "Stat": "Average",
                    },
                }
            ],
            "StartTime": datetime.datetime.now() - datetime.timedelta(hours=24),
            "EndTime": datetime.datetime.now(),
        }
        call = Common.generic_api_call(
            "cloudwatch",
            "get_metric_data",
            api_kwargs,
            "Retrieve metrics",
            "Cloudwatch",
            subcategory="Metrics",
            resource=f"{self.metric_namespace}/{self.metric_name}",
        )
        if call["Success"]:
            data = call["Response"]
            for idx in range(len(data["MetricDataResults"][0]["Timestamps"])):
                timestamp = data["MetricDataResults"][0]["Timestamps"][idx]
                value = data["MetricDataResults"][0]["Values"][idx]
                self.add_datapoint(timestamp, value)


class MetricLister(ResourceLister):
    """
    Lister resource for Cloudwatch Metrics.
    """

    prefix = "metric_list"
    title = "Metrics"

    resource_type = "metric"
    main_provider = "cloudwatch"
    category = "Cloudwatch"
    subcategory = "Metrics"
    list_method = "list_metrics"
    item_path = ".Metrics"
    columns = {
        "namespace": {"path": ".Namespace", "size": 16, "weight": 0, "sort_weight": 0},
        "name": {"path": ".MetricName", "size": 64, "weight": 1, "sort_weight": 1},
    }
    open_command = MetricViewer.opener
    open_selection_arg = "metric"

    def title_info(self):
        return f"{self.dimension[0]}={self.dimension[1]}"

    def __init__(
        self, *args, dimension, metric_namespace=None, metric_name=None, **kwargs
    ):
        self.dimension = dimension
        self.list_kwargs = {}
        if metric_namespace is not None:
            self.list_kwargs["Namespace"] = metric_namespace
        if metric_name is not None:
            self.list_kwargs["MetricName"] = metric_name
        super().__init__(*args, **kwargs)


class AlarmDescriber(Describer):
    """
    Describer resource for Cloudwatch Alarms.
    """

    prefix = "alarm_browser"
    title = "Cloudwatch Alarm"

    resource_type = "alarm"
    main_provider = "cloudwatch"
    category = "Cloudwatch"
    subcategory = "Alarm"
    describe_method = "describe_alarms"

    def __init__(self, *args, entry, **kwargs):
        self.resource_key = "cloudwatch"
        self.describe_kwargs = {
            "AlarmNames": [entry["name"]],
            "AlarmTypes": [f"{entry['type']}Alarm"],
        }
        self.object_path = (
            ".CompositeAlarms[0]"
            if entry["type"] == "Composite"
            else ".MetricAlarms[0]"
        )
        super().__init__(*args, entry=entry, **kwargs)


class AlarmLister(ResourceLister):
    """
    Lister resource for Cloudwatch Alarms.
    """

    prefix = "alarm_list"
    title = "Cloudwatch Alarms"
    command_palette = ["alarm", "cloudwatchalarm"]

    resource_type = "alarm"
    main_provider = "cloudwatch"
    category = "Cloudwatch"
    subcategory = "Alarms"
    list_method = "describe_alarms"
    item_path = ".CompositeAlarms + .MetricAlarms"
    columns = {
        "name": {"path": ".AlarmName", "size": 10, "weight": 0, "sort_weight": 0},
        "type": {
            "path": lambda x: "Metric"
            if "MetricName" in x or "Metrics" in x
            else "Composite",
            "size": 10,
            "weight": 1,
        },
        "actions": {
            "path": lambda x: "Enabled" if x["ActionsEnabled"] else "Disabled",
            "size": 10,
            "weight": 2,
        },
        "state": {"path": ".StateValue", "size": 10, "weight": 3},
        "description": {
            "path": ".AlarmDescription",
            "size": 64,
            "weight": 4,
        },
        "arn": {"path": ".AlarmArn", "hidden": True},
    }
    describe_command = AlarmDescriber.opener

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.confirm_template(
            "delete_alarms",
            {"AlarmNames": [SelectionAttribute("name")]},
            hotkey=ControlCodes.D,
            hotkey_tooltip="Delete alarm",
        )

    @ResourceLister.Autohotkey("e", "Edit", True)
    def edit(self, _):
        """
        Hotkey callback for editing a cloudwatch alarm.
        """
        editor = ListResourceDocumentEditor(
            "cloudwatch",
            "describe_alarms",
            ".CompositeAlarms[0]"
            if self.selection["type"] == "Composite"
            else ".MetricAlarms[0]",
            "put_composite_alarm"
            if self.selection["type"] == "Composite"
            else "put_metric_alarm",
            "AlarmNames",
            update_document_arg=None,
            entry_name_is_list=True,
            entry_key="name",
            entry_name_arg_update=None,
            as_json=True,
            message="Update successful",
            ignored_fields=[
                "AlarmArn",
                "AlarmConfigurationUpdatedTimestamp",
                "StateValue",
                "StateReason",
                "StateReasonData",
                "StateUpdatedTimestamp",
                "ActionsSuppressedReason",
                "ActionsSuppressedBy",
                "AlarmName",
            ],
            static_fields={"AlarmName": self.selection["name"]},
        )
        editor.edit(self.selection)


class CloudwatchAlarmDashboardBlock(KeyValueDashboardBlock):
    """
    Dashboard block for Cloudwatch Alarms.
    """

    description = "Cloudwatch Alarm summary"

    labels = {
        "total": "Total alarms",
        "OK": "Alarms in OK state",
        "ALARM": "Alarms in ALARM state",
        "INSUFFICIENT_DATA": "Alarms in INSUFFICIENT_DATA state",
    }
    thresholds = {
        "ALARM": [
            "dashboard_block_positive",
            1,
            "dashboard_block_negative",
        ],
        "INSUFFICIENT_DATA": [
            "dashboard_block_positive",
            1,
            "dashboard_block_neutral",
        ],
    }
    order = [
        "total",
        "OK",
        "ALARM",
        "INSUFFICIENT_DATA",
    ]

    def __init__(self, *args, **kwargs):
        self.alarm_data = {}
        super().__init__(*args, **kwargs)

    def refresh_data(self):
        resp = Common.generic_api_call(
            "cloudwatch",
            "describe_alarms",
            {},
            "Describe alarms",
            "Cloudwatch",
            subcategory="Alarm",
        )
        if resp["Success"]:
            asd = resp["Response"]
            info = {}
            info["total"] = len(asd["CompositeAlarms"]) + len(asd["MetricAlarms"])
            add_lines = []
            info["OK"] = 0
            info["ALARM"] = 0
            info["INSUFFICIENT_DATA"] = 0
            for alarm in asd["CompositeAlarms"]:
                info[alarm["StateValue"]] += 1
                if alarm["StateValue"] == "ALARM":
                    add_lines.append(
                        (
                            f"Composite: {alarm['AlarmName']}",
                            "dashboard_block_negative",
                            False,
                        )
                    )
            for alarm in asd["MetricAlarms"]:
                info[alarm["StateValue"]] += 1
                if alarm["StateValue"] == "ALARM":
                    add_lines.append(
                        (
                            f"MetricAlarm: {alarm['AlarmName']}",
                            "dashboard_block_negative",
                            False,
                        )
                    )
            if len(add_lines) > 0:
                add_lines.insert(
                    0, ("Alarms currently firing", "dashboard_block_label", True)
                )
            with self.mutex:
                self.additional_lines = add_lines
                self.alarm_data = asd
                self.info = info
                self.status = self.STATUS_READY
        else:
            with self.mutex:
                self.status = self.STATUS_ERROR
            return



"""
Module for informational display controls.
"""
from typing import Callable

from .termui.common import Commons, column_sizer
from .termui.control import Control
from .termui.dialog import DialogControl


class InfoDisplay(Control):
    """
    The info display is the control situated in the top left corner of the screen. It is intended to display information on the active session as a list
    of key-value pairs. The info display also registers the global hotkeys.

    Attributes
    ----------
    info : dict
        The actual info being displayed, as a mapping of label to data.
    order : list
        The order in which the info should be displayed.
    special_colors : dict
        A mapping of label to color for which pieces of data should be highlighted with a non-standard color, for example, for use with errors.
    cols : int
        The number of info columns to divide the available space to.
    highlight_color : awsc.termui.color.Color
        The color of the label.
    generic_color : awsc.termui.color.Color
        The default color of data.
    commander_hook : callable
        The callback to call for opening the commander.
    filterer_hook : callable
        The callback to call for opening the filterer.
    """

    def __init__(
        self,
        *args,
        info=None,
        cols=2,
        highlight_color=None,
        generic_color=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.info = {}
        self.order = info[:] if info is not None else []
        self.special_colors = {}
        for k in info:
            self.info[k] = None
        self.cols = cols
        self.highlight_color = highlight_color
        self.generic_color = generic_color
        self.commander_hook: Callable[[], None] = None
        self.filterer_hook: Callable[[], None] = None

    def input(self, key):
        if key == ":" and self.commander_hook is not None:
            # pylint: disable=not-callable # Member is callable or None, and check asserts that it is not None
            self.commander_hook()
            return True
        if key == "/" and self.filterer_hook is not None:
            # pylint: disable=not-callable # Member is callable or None, and check asserts that it is not None
            self.filterer_hook()
            return True
        return False

    def __getitem__(self, key):
        return self.info[key] if key in self.info else None

    def __setitem__(self, key, value):
        if key not in self.order:
            self.order.append(key)
        self.info[key] = value

    def paint(self):
        super().paint()
        (x0, x1), (y0, y1) = self.inner
        width = x1 - x0 + 1
        colw = int(width / self.cols)
        x = x0
        y = y0
        longest = column_sizer(y0, y1, self.order, self.info)
        col = 0

        for name in self.order:
            if name not in self.info:
                continue
            value = self.info[name]
            if value is None:
                value = ""
            display = name + ": "
            if len(display) < longest[col]:
                display += " " * (longest[col] - len(display))
            Commons.UIInstance.print(
                display, xy=(x, y), color=self.highlight_color, bold=True
            )
            text = value
            if len(text) > colw - longest[col]:
                text = text[: colw - longest[col]]
            Commons.UIInstance.print(
                text,
                xy=(x + longest[col], y),
                color=self.generic_color
                if name not in self.special_colors
                else self.special_colors[name],
            )
            y += 1
            if y > y1:
                (x, y, col) = (x + colw, y0, col + 1)


class NeutralDialog(DialogControl):
    """
    A dialog that does nothing and handles no inputs.
    """

    def input(self, key):
        return False



"""
Module for EBS volumes.
"""
from .base_control import Describer, ResourceLister, tagged_column_generator


def _ebs_determine_size(entry):
    """
    Column callback for returning the size of an EBS volume in a nice format.
    """
    return f"{entry['Size']} GiB"


def _ebs_determine_attachment(self, entry):
    """
    Column callback for figuring out which instance an EBS volume is attached to.
    """
    for attachment in entry["Attachments"]:
        if attachment["State"] in ("attaching", "attached", "busy"):
            return attachment["InstanceId"]
    return ""


class EBSDescriber(Describer):
    """
    Describer control for EBS volumes.
    """

    prefix = "ebs_browser"
    title = "EBS Volume"

    resource_type = "ebs volume"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "EBS Volume"
    describe_method = "describe_volumes"
    describe_kwarg_name = "VolumeIds"
    describe_kwarg_is_list = True
    object_path = ".Volumes[0]"
    default_entry_key = "id"


class EBSResourceLister(ResourceLister):
    """
    Lister control for EBS volumes.
    """

    prefix = "ebs_list"
    title = "EBS Volumes"
    command_palette = ["ebs"]

    resource_type = "ebs volume"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "EBS Volume"
    list_method = "describe_volumes"
    item_path = ".Volumes"
    columns = {
        "id": {"path": ".VolumeId", "size": 20, "weight": 0},
        **tagged_column_generator("name", "name", weight=1, sort_weight=0, size=30),
        "size": {"path": _ebs_determine_size, "size": 10, "weight": 2},
        "state": {"path": ".State", "size": 10, "weight": 3},
        "type": {"path": ".VolumeType", "size": 10, "weight": 4},
        "attached to": {"path": _ebs_determine_attachment, "size": 16, "weight": 5},
    }
    primary_key = "id"
    describe_command = EBSDescriber.opener



"""
Module for VPC resources.
"""
from .base_control import Describer, ResourceLister, tagged_column_generator


class VPCDescriber(Describer):
    """
    Describer control for VPC resources.
    """

    prefix = "vpc_browser"
    title = "VPC"

    resource_type = "vpc"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "VPC"
    describe_method = "describe_vpcs"
    describe_kwarg_name = "VpcIds"
    describe_kwarg_is_list = True
    object_path = ".Vpcs[0]"
    default_entry_key = "id"


def _vpc_determine_default(self, vpc, *args):
    if vpc["IsDefault"]:
        return "✓"
    return ""


class VPCResourceLister(ResourceLister):
    """
    Lister control for VPC resources.
    """

    prefix = "vpc_list"
    title = "VPCs"
    command_palette = ["vpc"]

    resource_type = "vpc"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "VPC"
    list_method = "describe_vpcs"
    item_path = ".Vpcs"
    columns = {
        "id": {
            "path": ".VpcId",
            "size": 30,
            "weight": 0,
            "sort_weight": 1,
        },
        **tagged_column_generator("name", "name", weight=1, sort_weight=0, size=30),
        "default": {
            "path": _vpc_determine_default,
            "size": 3,
            "weight": 2,
        },
        "cidr": {
            "path": ".CidrBlock",
            "size": 18,
            "weight": 3,
        },
        "state": {
            "path": ".State",
            "size": 9,
            "weight": 4,
        },
        # "arn": {"path": ".SubnetArn", "hidden": True},
    }
    describe_command = VPCDescriber.opener
    open_command = "t"
    primary_key = "id"



"""
Module for hotkey display.
"""

from .common import Common
from .termui.alignment import Dimension, TopRightAnchor
from .termui.common import column_sizer
from .termui.control import Control
from .termui.ui import ControlCodes


class HotkeyDisplay(Control):
    """
    The hotkey display is the display element in the top right, listing valid hotkeys accepted by the current controls on screen. Each element will provider its
    own hotkey display instance, filled with its own and the global hotkeys.

    Attributes
    ----------
    translations : dict
        A mapping of key names to their nicely presentable forms.
    holder : awsc.termui.control.HotkeyControl
        The control whose hotkeys this displays.
    cols : int
        The number of info columns to divide the available space to.
    highlight_color : awsc.termui.color.Color
        The color of the keys on the hotkey display.
    generic_color : awsc.termui.color.Color
        The color of the tooltips on the hotkey display.
    session : awsc.session.Session
        A reference to the session object.
    """

    @classmethod
    def opener(cls, *args, caller, **kwargs):
        """
        Simplified initializer for a hotkey display, following the patterns defined by other controls.
        """

        return cls(
            caller.parent,
            TopRightAnchor(1, 0),
            Dimension("33%|50", 8),
            holder=caller,
            session=Common.Session,
            highlight_color=Common.color("hotkey_display_title"),
            generic_color=Common.color("hotkey_display_value"),
        )

    translations = {
        "KEY_DOWN": "↓",
        "KEY_RIGHT": "→",
        "KEY_UP": "↑",
        "KEY_LEFT": "←",
        "KEY_ENTER": "↲",
        "\n": "↲",
        "\t": "↹",
        "KEY_BACKSPACE": "⇦",
        "KEY_DELETE": "del",
        "KEY_END": "end",
        "KEY_HOME": "home",
        "KEY_ESCAPE": "esc",
        "KEY_PGUP": "pgup",
        "KEY_PGDOWN": "pgdn",
        "KEY_INSERT": "ins",
        ControlCodes.A: "ctrl-a",
        ControlCodes.B: "ctrl-b",
        ControlCodes.C: "ctrl-c",
        ControlCodes.D: "ctrl-d",
        ControlCodes.E: "ctrl-e",
        ControlCodes.F: "ctrl-f",
        ControlCodes.G: "ctrl-g",
        ControlCodes.K: "ctrl-k",
        ControlCodes.L: "ctrl-l",
        ControlCodes.M: "ctrl-m",
        ControlCodes.N: "ctrl-n",
        ControlCodes.O: "ctrl-o",
        ControlCodes.P: "ctrl-p",
        ControlCodes.Q: "ctrl-q",
        ControlCodes.R: "ctrl-r",
        ControlCodes.S: "ctrl-s",
        ControlCodes.T: "ctrl-t",
        ControlCodes.U: "ctrl-u",
        ControlCodes.V: "ctrl-v",
        ControlCodes.W: "ctrl-w",
        ControlCodes.X: "ctrl-x",
        ControlCodes.Y: "ctrl-y",
        ControlCodes.Z: "ctrl-z",
    }

    def __init__(
        self,
        *args,
        holder,
        session=None,
        cols=2,
        highlight_color=None,
        generic_color=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.holder = holder
        self.cols = cols
        self.highlight_color = highlight_color
        self.generic_color = generic_color
        self.session = session

    def paint(self):
        super().paint()
        ((x0, x1), (y0, y1)) = self.inner
        width = x1 - x0 + 1
        colw = int(width / self.cols)
        x = x0
        y = y0
        col = 0
        longest = []
        tooltips = {**self.holder.tooltips, **self.session.global_hotkey_tooltips}
        labels = []
        generated = {}
        for hotkey, tooltip in tooltips.items():
            display = (
                "<"
                + (
                    HotkeyDisplay.translations[hotkey]
                    if hotkey in HotkeyDisplay.translations
                    else hotkey
                )
                + "> "
            )
            labels.append(display)
            generated[hotkey] = display
        longest = column_sizer(y0, y1, labels, None)
        for hotkey, tooltip in tooltips.items():
            display = generated[hotkey]
            if len(display) < longest[col]:
                display += " " * (longest[col] - len(display))
            Common.Session.ui.print(
                display, xy=(x, y), color=self.highlight_color, bold=True
            )
            text = tooltip if not callable(tooltip) else tooltip()

            if len(text) > colw - longest[col]:
                text = text[: colw - longest[col]]
            Common.Session.ui.print(
                text, xy=(x + longest[col], y), color=self.generic_color
            )
            y += 1
            if y > y1:
                (x, y, col) = (x + colw, y0, col + 1)



"""
This module defines a bar graph control that can be used to display timeseries.
"""
import datetime

from .color import ColorDarkGreen, ColorGold
from .common import Commons
from .control import HotkeyControl


class BarGraph(HotkeyControl):
    """
    In the most particular culmination of horrible ideas, this control attempts to - remarkably successfully, if I might add - display a bar
    graph in a terminal environment. It is fully capable of displaying a timestamped time series in an understandable fashion.

    Attributes
    ----------
    DPB_ZeroOrLessToMax : int
        Data point behaviour constant. Bar graphs with this data point behaviour will extend their min-point to the negative numbers if any
        datapoint is negative, but will never place their min-point higher than zero.
    DPB_ZeroToMax : int
        Data point behaviour constant. Bar graphs with this data point behaviour will always place their min-point at zero, even if some
        datapoints are negative. Negative datapoints will be displayed as an empty bar.
    DPB_MinToMax : int
        Data point behaviour constant. Bar graphs with this data point behaviour will always place their min-point at the value of the lowest
        datapoint, regardless of whether it is negative, zero or positive.
    subdivisions : dict
        The characters used to represent subdivisions of a single character cell. The key zero represents an empty cell, positive numbers up
        to 8 represent that many eighths of a cell to be filled on a bar extending upwards, while negative numbers are the same down to -8
        for a bar extending downwards. The question mark represents cells where something went horribly wrong.
    datapoints : list(tuple(datetime.datetime, float))
        A list of datapoints in the time series. Each datapoint is a tuple of the time of measuring, and the value of the datapoint.
    max_point : float
        The calculated highest datapoint in the graph.
    min_point : float
        The calculated lowest datapoint in the graph.
    color : awsc.termui.color.Color
        The main color to use for text and unselected columns.
    highlight_color : awsc.termui.color.Color
        The color to use for highlighted columns.
    left : int
        The offset for how many columns to skip before drawing - basically, the amount you are scrolled to the right by.
    highlight : int
        The index of the selected column.
    """

    DPB_ZeroOrLessToMax = 0
    DPB_ZeroToMax = 1
    DPB_MinToMax = 2

    def __init__(
        self,
        *args,
        color=ColorGold,
        highlight_color=ColorDarkGreen,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.subdivisions = {
            0: " ",
            1: "▁",
            2: "▂",
            3: "▃",
            4: "▄",
            5: "▅",
            6: "▆",
            7: "▇",
            8: "█",
            -1: "▔",
            -2: "▔",
            -3: "▔",
            -4: "▀",
            -5: "▀",
            -6: "▀",
            -7: "▀",
            -8: "█",
            "?": "░",
        }

        self._data_points_behaviour = BarGraph.DPB_ZeroOrLessToMax

        self.datapoints = []
        self.max_point = 0
        self.min_point = 0
        now = datetime.datetime.now()
        self.min_time = now
        self.max_time = now
        self.color = color
        self.highlight_color = highlight_color
        self.add_hotkey("KEY_END", self.end)
        self.add_hotkey("KEY_HOME", self.home)
        self.add_hotkey("KEY_LEFT", self.scroll_left)
        self.add_hotkey("KEY_RIGHT", self.scroll_right)

        self.left = 0
        self.highlight = 0

    def scroll_left(self, *args):
        """
        Hotkey callback for KEY_LEFT. Moves the selection one column to the left.
        """
        if self.highlight > 0:
            self.highlight -= 1
            if self.left > self.highlight:
                self.left = self.highlight
            Commons.UIInstance.dirty = True

    def home(self, *args):
        """
        Hotkey callback for KEY_HOME. Moves the selection all the way to the left.
        """
        self.left = 0
        self.highlight = 0
        Commons.UIInstance.dirty = True

    def end(self, *args):
        """
        Hotkey callback for KEY_END. Moves the selection all the way to the right.
        """
        self.highlight = len(self.datapoints) - 1
        self.left = max(len(self.datapoints) - self.colspace, 0)
        Commons.UIInstance.dirty = True

    def scroll_right(self, *args):
        """
        Hotkey callback for KEY_RIGHT. Moves the selection one column to the right.
        """
        if self.highlight < len(self.datapoints) - 1:
            self.highlight += 1
            if self.highlight >= self.left + self.colspace:
                self.left += 1
            Commons.UIInstance.dirty = True

    @property
    def data_points_behaviour(self):
        """
        Property representing the data points behaviour of this bar graph. Can be set to one of the data points behaviour constants defined
        in this class. Controls how the minimum point on the graph is determined when painting the graph.

        Raises
        ------
        ValueError
            The datapoint behaviour being set is not one of the allowed values.
        """
        return self._data_points_behaviour

    @data_points_behaviour.setter
    def data_points_behaviour(self, value):
        self._data_points_behaviour = value
        if value == BarGraph.DPB_ZeroOrLessToMax:
            self.min_point = 0
            for datapoint in self.datapoints:
                if datapoint[1] < self.min_point:
                    self.min_point = datapoint[1]
        elif value == BarGraph.DPB_ZeroToMax:
            self.min_point = 0
        elif value == BarGraph.DPB_MinToMax:
            self.min_point = self.max_point
            for datapoint in self.datapoints:
                if datapoint[1] < self.min_point:
                    self.min_point = datapoint[1]
        else:
            raise ValueError

    def add_datapoint(self, timestamp, datapoint):
        """
        Inserts a new datapoint into the bar graph.

        Parameters
        ----------
        timestamp : datetime.datetime
            The time of measurement for the datapoint.
        datapoint : float
            The value at the time of the measurement.
        """
        try:
            if len(self.datapoints) == 0:
                self.max_point = datapoint
                if self._data_points_behaviour == BarGraph.DPB_ZeroOrLessToMax:
                    self.min_point = min(0, datapoint)
                elif self._data_points_behaviour == BarGraph.DPB_ZeroToMax:
                    self.min_point = 0
                elif self._data_points_behaviour == BarGraph.DPB_MinToMax:
                    self.min_point = datapoint
            else:
                self.max_point = max(self.max_point, datapoint)
                if self._data_points_behaviour == BarGraph.DPB_ZeroOrLessToMax:
                    self.min_point = min(self.min_point, 0, datapoint)
                elif self._data_points_behaviour == BarGraph.DPB_ZeroToMax:
                    self.min_point = 0
                elif self._data_points_behaviour == BarGraph.DPB_MinToMax:
                    self.min_point = min(self.min_point, datapoint)
            for idx, existing_datapoint in enumerate(self.datapoints):
                existing_timestamp = existing_datapoint[0]
                if existing_timestamp > timestamp:
                    self.datapoints.insert(idx, (timestamp, datapoint))
                    return
            self.datapoints.append((timestamp, datapoint))
        finally:
            self.left = max(len(self.datapoints) - self.colspace, 0)
            self.highlight = len(self.datapoints) - 1
            Commons.UIInstance.dirty = True

    @HotkeyControl.border.setter  # type: ignore[attr-defined]
    def border(self, value):
        self._border = value
        # Do not execute on border initialization
        if hasattr(self, "datapoints"):
            self.left = max(len(self.datapoints) - self.colspace, 0)
            Commons.UIInstance.dirty = True

    @property
    def colspace(self):
        """
        Read-only property representing the amount of displayable columns. This is shorter than the inner width of the control as there
        is a label for the minimum and maximum values on the graph.
        """
        corners = self.corners
        wspace = corners[0][1] - corners[0][0] + 1 - (0 if self.border is None else 2)
        max_label = f"{float(self.max_point):.2f}"
        min_label = f"{float(self.min_point):.2f}"
        x_label_space = max(len(max_label), len(min_label)) + 1
        return wspace - x_label_space

    def paint(self):
        super().paint()
        corners = self.corners
        if len(self.datapoints) == 0:
            Commons.UIInstance.print_centered(
                "No datapoints loaded", self.corners, color=self.color, bold=True
            )
            return
        wspace = corners[0][1] - corners[0][0] + 1 - (0 if self.border is None else 2)
        hspace = corners[1][1] - corners[1][0] + 1 - (0 if self.border is None else 2)
        x0 = corners[0][0] + (0 if self.border is None else 1)
        y0 = corners[1][0] + (0 if self.border is None else 1)

        max_label = f"{float(self.max_point):.2f}"
        min_label = f"{float(self.min_point):.2f}"

        x_label_space = max(len(max_label), len(min_label)) + 1
        colspace = wspace - x_label_space
        if colspace == 0:
            return
        first_col = self.left
        last_col = min(self.left + colspace - 1, len(self.datapoints) - 1)
        col_width = float(colspace) / float((last_col - first_col + 1))
        roll = col_width - int(col_width)
        col_width = int(col_width)
        rolling = 0

        rowspace = hspace - 2
        if rowspace == 0:
            return
        point_step = (self.max_point - self.min_point) / rowspace

        xb0 = x0 + x_label_space

        Commons.UIInstance.print(max_label, (x0, y0), color=self.color)
        Commons.UIInstance.print(min_label, (x0, y0 + rowspace - 1), color=self.color)
        Commons.UIInstance.print(
            self.datapoints[first_col][0].isoformat(),
            (xb0, y0 + rowspace),
            color=self.color,
        )
        last_ts = self.datapoints[last_col][0].isoformat()
        Commons.UIInstance.print(
            last_ts,
            (xb0 + colspace - len(last_ts) - 1, y0 + rowspace),
            color=self.color,
        )
        Commons.UIInstance.print(
            f"Value at {self.datapoints[self.highlight][0].isoformat()}: {self.datapoints[self.highlight][1]:.2f}",
            (xb0, y0 + rowspace + 1),
            color=self.highlight_color,
        )

        value = self.max_point
        for y in range(y0, y0 + rowspace):
            row_bounds = (value - point_step, value)
            value -= point_step
            row = ["", "", ""]
            rowindex = 0
            rolling = 0
            for col in range(first_col, last_col + 1):
                datapoint = self.datapoints[col][1]
                colw = col_width
                if col == self.highlight:
                    rowindex = 1
                if rolling > 1:
                    colw += 1
                    rolling -= 1
                if datapoint < row_bounds[0]:
                    if row_bounds[0] > 0:
                        row[rowindex] += self.subdivisions[0] * colw
                    else:
                        row[rowindex] += self.subdivisions[-8] * colw
                elif datapoint > row_bounds[1]:
                    if row_bounds[1] > 0:
                        row[rowindex] += self.subdivisions[8] * colw
                    else:
                        row[rowindex] += self.subdivisions[0] * colw
                else:
                    if row_bounds[0] < 0 < row_bounds[1]:
                        row[rowindex] += self.subdivisions[8] * colw
                    elif datapoint > 0:
                        if point_step != 0:
                            subdivision = int(
                                (float(datapoint - row_bounds[0]) / point_step) * 8
                            )
                        else:
                            subdivision = "?"
                        row[rowindex] += self.subdivisions[subdivision] * colw
                    else:
                        if point_step != 0:
                            subdivision = int(
                                (float(datapoint - row_bounds[1]) / point_step) * 8
                            )
                        else:
                            subdivision = "?"
                        row[rowindex] += self.subdivisions[subdivision] * colw
                rolling += roll
                if col == self.highlight:
                    rowindex = 2
            Commons.UIInstance.print(row[0], (xb0, y), color=self.color)
            if len(row[1]) > 0:
                Commons.UIInstance.print(
                    row[1], (xb0 + len(row[0]), y), color=self.highlight_color
                )
                if len(row[2]) > 0:
                    Commons.UIInstance.print(
                        row[2], (xb0 + len(row[0]) + len(row[1]), y), color=self.color
                    )



"""
Module for S3 resources.
"""
import shutil
from pathlib import Path

from .base_control import Describer, GenericDescriber, ResourceLister
from .common import Common
from .termui.list_control import ListEntry
from .termui.ui import ControlCodes

_s3_icons = {
    "folder": "🗀",
    "default": "🗈",
    "ext.txt": "🖹",
    "ext.png": "🖻",
    "ext.jpg": "🖻",
    "ext.gif": "🖻",
    "ext.jpeg": "🖻",
    "ext.svg": "🖻",
    "ext.doc": "🖺",
    "ext.docx": "🖺",
    "ext.xls": "🖺",
    "ext.xlsx": "🖺",
    "ext.ppt": "🖺",
    "ext.pptx": "🖺",
}


def _s3_object_determine_path(self, entry, *args):
    return entry["Key"] if "Key" in entry else entry["Prefix"]


def _s3_object_determine_name(self, entry, *args):
    if "Prefix" in entry:
        return entry["Prefix"].strip("/").split("/")[-1]
    return entry["Key"].split("/")[-1]


def _s3_object_determine_is_dir(self, entry, *args):
    return "/" if "Prefix" in entry else ""


def _s3_object_determine_size(self, entry, *args):
    if "Prefix" in entry:
        return ""
    b_prefix = ["", "Ki", "Mi", "Gi", "Ti", "Ei"]
    b_idx = 0
    size = float(entry["Size"])
    while size >= 1024:
        b_idx += 1
        size /= 1024
        if b_idx == len(b_prefix) - 1:
            break
    return f"{size:.2f} {b_prefix[b_idx]}B"


def _s3_object_determine_icon(self, entry, *args):
    if "Prefix" in entry:
        return _s3_icons["folder"]
    ext = entry["Key"].split("/")[-1].split(".")[-1]
    ext_key = f"ext.{ext}"
    if ext_key in _s3_icons:
        return _s3_icons[ext_key]
    return _s3_icons["default"]


class S3ObjectLister(ResourceLister):
    """
    Lister control for S3 objects.

    Attributes
    ----------
    bucket : str
        Name of the bucket being browsed.
    path : str
        Initial path within the bucket.
    """

    prefix = "s3_object_list"
    title = "S3 Browser"

    resource_type = "s3 object"
    main_provider = "s3"
    category = "S3"
    subcategory = "Object"
    list_method = "list_objects"
    item_path = ".Contents + .CommonPrefixes"
    columns = {
        "icon": {
            "path": _s3_object_determine_icon,
            "size": 1,
            "weight": 0,
        },
        "name": {
            "path": _s3_object_determine_name,
            "size": 150,
            "weight": 1,
            "sort_weight": 1,
        },
        "gateway": {
            "path": _s3_object_determine_size,
            "size": 30,
            "weight": 2,
        },
        "is_dir": {
            "path": _s3_object_determine_is_dir,
            "hidden": True,
            "sort_weight": 0,
        },
        "size_in_bytes": {
            "path": ".Size",
            "hidden": True,
        },
        "path": {
            "path": _s3_object_determine_path,
            "hidden": True,
        },
    }
    primary_key = "path"

    def title_info(self):
        if self.path is None:
            return self.bucket
        return f"{self.bucket}/{self.path}"

    def __init__(self, *args, bucket, path=None, **kwargs):
        if isinstance(bucket, ListEntry):
            self.bucket = bucket["name"]
        else:
            self.bucket = bucket
        self.path = path
        self.list_kwargs = {"Bucket": self.bucket, "MaxKeys": 100, "Delimiter": "/"}
        self.next_marker = "NextMarker"
        self.next_marker_arg = "Marker"
        if self.path is not None:
            self.list_kwargs["Prefix"] = self.path
        self.open_command = S3ObjectLister.opener
        self.open_selection_arg = "path"
        super().__init__(*args, **kwargs)
        self.add_hotkey(ControlCodes.D, self.empty, "Cancel download")

    def open(self, *args):
        if (
            self.open_command is not None
            and self.selection is not None
            and self.selection["is_dir"] == "/"
        ):
            subpath = self.get_selection_path()
            self.command(
                self.open_command,
                **{
                    self.open_selection_arg: subpath,
                    "bucket": self.bucket,
                    "pushed": True,
                },
            )

    def get_selection_path(self):
        """
        Gets the full path for the current selection within the bucket.

        Returns
        -------
        str
            The full path of the current selection. Has a trailing slash if it's a directory.
        """
        if self.selection is not None:
            if self.path is None:
                return f"{self.selection['name']}{self.selection['is_dir']}"
            return f"{self.path}{self.selection['name']}{self.selection['is_dir']}"
        return None

    def get_cache_name(self):
        """
        Generates a cache name for the currently selected object.

        Returns
        -------
        str
            The cache name for the file.
        """
        return f"{self.bucket}/{self.get_selection_path()}".replace("/", "--")

    def cache_fetch(self, obj, *args, **kwargs):
        """
        Fetches an S3 object to local cache.

        Parameters
        ----------
        obj : str
            The name of the object to fetch.

        Returns
        -------
        str or bytes
            The contents of the file.
        """
        selection_path = self.get_selection_path()
        obj_size = float(self.selection["size_in_bytes"])
        downloaded = 0

        def fn(chunk):
            nonlocal downloaded
            downloaded += chunk
            perc = float(downloaded) / obj_size
            key = Common.Session.ui.check_one_key()
            if key == ControlCodes.C:
                raise KeyboardInterrupt
            if key == ControlCodes.D:
                raise CancelDownload
            Common.Session.ui.progress_bar_paint(perc)

        Common.generic_api_call(
            "s3",
            "download_file",
            {
                "Bucket": self.bucket,
                "Key": selection_path,
                "Filename": obj,
                "Callback": fn,
            },
            "Download File",
            "S3",
            subcategory="Object",
            resource=f"{self.bucket}/{selection_path}",
        )

        return Common.Session.ui.read_file(obj)

    @ResourceLister.Autohotkey("d", "Download", True)
    def download_selection(self, *args, **kwargs):
        """
        Hotkey callback for downloading an S3 object.
        """
        if self.selection["is_dir"] == "/":
            Common.Session.set_message(
                "Cannot download directory as file.", Common.color("message_info")
            )
            return
        path = Path.home() / "Downloads" / "awsc"
        path.mkdir(parents=True, exist_ok=True)
        try:
            objpath = Common.Session.ui.filecache_path(
                self.get_cache_name(), self.cache_fetch
            )
        except CancelDownload:
            Common.Session.set_message(
                "Download cancelled", Common.color("message_info")
            )
            return

        destpath = path / self.selection["name"]
        shutil.copy(objpath, destpath)
        Common.Session.set_message(
            f"Downloaded file to {destpath.resolve()}",
            Common.color("message_success"),
        )

    @ResourceLister.Autohotkey("v", "View", True)
    def view_selection(self, *args, **kwargs):
        """
        Hotkey callback for viewing an S3 object as text.
        """
        if self.selection is not None:
            if self.selection["is_dir"] == "/":
                Common.Session.set_message(
                    "Cannot display directory as text file.",
                    Common.color("message_info"),
                )
                return None
            try:
                data = Common.Session.ui.filecache(
                    self.get_cache_name(), self.cache_fetch
                )
            except CancelDownload:
                Common.Session.set_message(
                    "Download cancelled", Common.color("message_info")
                )
                return None
            if isinstance(data, bytes):
                data = "<binary>"
            Common.Session.push_frame(
                GenericDescriber.opener(
                    **{
                        "describing": f"File: {self.get_selection_path()}",
                        "content": data,
                        "pushed": True,
                    }
                )
            )
        return None


class S3Describer(Describer):
    """
    Describer control for S3 buckets.
    """

    prefix = "s3_browser"
    title = "S3 Bucket"

    resource_type = "bucket"
    main_provider = "s3"
    category = "S3"
    subcategory = "Bucket"
    describe_method = "get_bucket_policy"
    describe_kwarg_name = "Bucket"
    object_path = "."


def _s3_determine_location(self, entry):
    loc_resps = Common.generic_api_call(
        "s3",
        "get_bucket_location",
        {"Bucket": entry["name"]},
        "Get Bucket Location",
        "S3",
        subcategory="Bucket",
        resource=entry["name"],
    )
    if loc_resps["Success"]:
        loc_resp = loc_resps["Response"]
        return loc_resp["LocationConstraint"]
    return "<n/a>"


class S3ResourceLister(ResourceLister):
    """
    Lister control for S3 buckets.
    """

    prefix = "s3_list"
    title = "S3 Buckets"
    command_palette = ["s3"]

    def __init__(self, *args, **kwargs):
        self.resource_key = "s3"
        self.list_method = "list_buckets"
        self.item_path = ".Buckets"
        self.column_paths = {
            "name": ".Name",
            "location": _s3_determine_location,
        }
        self.imported_column_sizes = {
            "name": 60,
            "location": 20,
        }
        self.describe_command = S3Describer.opener
        self.open_command = S3ObjectLister.opener
        self.open_selection_arg = "bucket"
        self.primary_key = "name"

        self.imported_column_order = ["name", "location"]
        self.sort_column = "name"
        super().__init__(*args, **kwargs)


class CancelDownload(Exception):
    """
    Exception class thrown to abort the ongoing download.
    """



"""
Module for SQS resources.
"""
from typing import Dict

from .base_control import (
    Describer,
    ListResourceDocumentCreator,
    ResourceLister,
    SelectionAttribute,
    TemplateDict,
)
from .common import Common
from .termui.ui import ControlCodes


def _sqs_determine_queue_arn(result):
    return SQSLister.get_attrib_cache(result, "QueueArn")


def _sqs_determine_available_messages(result):
    return SQSLister.get_attrib_cache(result, "ApproximateNumberOfMessages")


class SQSDescriber(Describer):
    """
    Describer control for SQS queues.
    """

    prefix = "sqs_browser"
    title = "Queue"

    resource_type = "queue"
    main_provider = "sqs"
    category = "SQS"
    subcategory = "Queue"
    describe_method = "get_queue_attributes"
    describe_kwarg_name = "QueueUrl"
    object_path = "."
    describe_kwargs_override = {"AttributeNames": ["All"]}
    default_entry_key = "url"


class SQSLister(ResourceLister):
    """
    Lister control for SQS queues.
    """

    _sqs_attrib_cache: Dict[str, str] = {}

    prefix = "sqs_list"
    title = "Queues"
    command_palette = ["sqs", "queue", "queues"]

    resource_type = "Queue"
    main_provider = "sqs"
    category = "SQS"
    subcategory = "Queue"
    list_method = "list_queues"
    item_path = ".QueueUrls"
    columns = {
        "url": {
            "path": ".",
            "size": 80,
            "weight": 0,
            "sort_weight": 0,
        },
        "available": {
            "path": _sqs_determine_available_messages,
            "size": 9,
            "weight": 1,
        },
        "name": {"path": ".", "hidden": True},
        "arn": {"path": _sqs_determine_queue_arn, "hidden": True},
    }
    describe_command = SQSDescriber.opener
    primary_key = "url"

    @ResourceLister.Autohotkey("s", "Send message", True)
    def send_message(self, _):
        """
        Hotkey callback for inserting a message into the queue.
        """
        if self.selection is None:
            return
        doc = {
            "MessageBody": "",
            "DelaySeconds": 0,
            "MessageAttributes": {},
            "MessageSystemAttributes": {},
        }
        if self.selection["arn"].endswith(".fifo"):
            doc["MessageDeduplicationId"]: ""
            doc["MessageGroupId"]: ""
        creator = ListResourceDocumentCreator(
            "sqs",
            "send_message",
            None,
            as_json=["MessageAttributes", "MessageSystemAttributes"],
            initial_document=doc,
            static_fields={"QueueUrl": self.selection["url"]},
            message="Message successfully sent",
        )
        creator.edit()

    @ResourceLister.Autohotkey("n", "Create queue")
    def create_queue(self, _):
        """
        Hotkey callback for creating a new queue.
        """
        doc = {
            "QueueName": "",
            "Attributes": {
                "DelaySeconds": "0",
                "MaximumMessageSize": "262144",
                "MessageRetentionPeriod": "345600",
                "Policy": "",
                "ReceiveMessageWaitTimeSeconds": "0",
                "RedrivePolicy": "",
                "VisibilityTimeout": "30",
                "KmsMasterKeyId": "",
                "KmsDataKeyReusePeriodSeconds": "300",
                "SqsManagedSseEnabled": "false",
                "FifoQueue": "false",
                "ContentBasedDeduplication": "false",
                "DeduplicationScope": "queue",
                "FifoThroughputLimit": "perQueue",
            },
            "tags": {},
        }
        creator = ListResourceDocumentCreator(
            "sqs",
            "create_queue",
            None,
            as_json=["Attributes", "tags"],
            initial_document=doc,
        )
        creator.edit()

    @ResourceLister.Autohotkey(ControlCodes.P, "Purge Queue", True)
    def purge(self, _):
        """
        Hotkey callback for purging a queue.
        """
        self.confirm_template(
            "purge_queue",
            TemplateDict(
                {
                    "QueueUrl": SelectionAttribute("url"),
                }
            ),
            action_name="Purge",
        )(self.selection)

    @ResourceLister.Autohotkey(ControlCodes.D, "Delete Queue", True)
    def delete_queue(self, _):
        """
        Hotkey callback for queue deletion.
        """
        self.confirm_template(
            "delete_queue",
            TemplateDict(
                {
                    "QueueUrl": SelectionAttribute("url"),
                }
            ),
        )(self.selection)

    def refresh_data(self, *args, **kwargs):
        SQSLister._attrib_cache = {}
        return super().refresh_data(*args, **kwargs)

    @classmethod
    def get_attrib_cache(cls, url, attrib):
        """
        Returns a queue attribute from the attribute cache.

        Parameters
        ----------
        url : str
            The url of the queue for which an attribute is being queried.
        attrib : str
            The name of the attribute to query.

        Returns
        -------
        str
            The queue attribute, or n/a if it does not exist.
        """
        if url not in cls._attrib_cache:
            resps = Common.generic_api_call(
                "sqs",
                "get_queue_attributes",
                {"QueueUrl": url, "AttributeNames": ["All"]},
                "Get Queue Attributes",
                "SQS",
                subcategory="Queue",
                resource=url,
            )
            if resps["Success"]:
                attribs = resps["Response"]
                cls._attrib_cache[url] = attribs
        try:
            return cls._attrib_cache[url]["Attributes"][attrib]
        except KeyError:
            return "<n/a>"



"""
Module for subnet resources.
"""
from .base_control import Describer, ResourceLister, tagged_column_generator
from .common import Common


class SubnetDescriber(Describer):
    """
    Describer control for subnet resources.
    """

    prefix = "subnet_browser"
    title = "Subnet"

    resource_type = "subnet"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "Subnet"
    describe_method = "describe_subnets"
    describe_kwarg_name = "SubnetIds"
    describe_kwarg_is_list = True
    object_path = ".Subnets[0]"
    default_entry_key = "id"


def _subnet_determine_public(self, subnet, *args):
    if subnet["MapPublicIpOnLaunch"]:
        return "✓"
    return ""


@ResourceLister.Autocommand("VPCResourceLister", "t", "View Subnets", "vpc")
@ResourceLister.Autocommand(
    "DBSubnetGroupResourceLister", "t", "View Subnets", "db_subnet_group"
)
class SubnetResourceLister(ResourceLister):
    """
    Lister control for subnet resources.
    """

    prefix = "subnet_list"
    title = "Subnets"
    command_palette = ["subnet"]

    resource_type = "Subnet"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "Subnet"
    list_method = "describe_subnets"
    item_path = ".Subnets"
    columns = {
        "id": {"path": ".SubnetId", "size": 30, "weight": 0},
        **tagged_column_generator("name", "name", weight=1, sort_weight=2, size=30),
        "vpc": {"path": ".VpcId", "size": 30, "weight": 2, "sort_weight": 0},
        "cidr": {"path": ".CidrBlock", "size": 18, "weight": 3, "sort_weight": 1},
        "AZ": {"path": ".AvailabilityZone", "size": 20, "weight": 4},
        "public": {"path": _subnet_determine_public, "size": 3, "weight": 5},
        "arn": {"path": ".SubnetArn", "hidden": True},
    }
    describe_command = SubnetDescriber.opener
    primary_key = "id"

    def title_info(self):
        if self.vpc is not None:
            return self.vpc["id"]
        if self.db_subnet_group is not None:
            return self.db_subnet_group["name"]
        return None

    def __init__(self, *args, vpc=None, db_subnet_group=None, **kwargs):
        self.vpc = vpc
        self.db_subnet_group = db_subnet_group
        if vpc is not None:
            self.list_kwargs = {
                "Filters": [
                    {
                        "Name": "vpc-id",
                        "Values": [
                            self.vpc["id"],
                        ],
                    }
                ]
            }
        if db_subnet_group is not None:
            self.list_kwargs = self.get_db_subnet_ids()
        super().__init__(*args, **kwargs)

    def get_db_subnet_ids(self, *args):
        """
        Generates a filetr for listing subnets based on a DB subnet group.

        Returns
        -------
        dict
            A filter to pass to list_kwargs for filtering for the DB subnet group in self.db_subnet_group.
        """
        dsg_resp = Common.generic_api_call(
            "rds",
            "describe_db_subnet_groups",
            {"DBSubnetGroupName": self.db_subnet_group.name},
            "Describe DB Subnet Group",
            "RDS",
            subcategory="Subnet Group",
            resource=self.db_subnet_group.name,
        )
        if not dsg_resp["Success"]:
            return {"SubnetIds": []}
        dsg = dsg_resp["Response"]
        return {
            "SubnetIds": [
                s["SubnetIdentifier"] for s in dsg["DBSubnetGroups"][0]["Subnets"]
            ]
        }



"""
This module defines screen buffer classes for buffering output to the terminal.
"""


class Character:
    """
    Represents a single character cell on the screen. Each cell has a character value, a color, and a boldness flag.

    Attributes
    ----------
    value : str
        The character to display in the cell.
    color : awsc.termui.color.Color
        A callable which takes a string and returns a string. Designed with the Color object in mind, but can be replaced with any callable.
        Called when the cell is reset or changed to color the display character.
    bold : bool
        Whether the character should be displayed in bold.
    out : str
        The calculated output of the cell. Contains a control code to color the character, then the character, then a reset control code.
    dirty : bool
        Whether the character output needs recalculating.
    """

    def __init__(self):
        """
        Initializes a Character object.
        """
        self.value = " "
        self.color = Character.nullcolor
        self.bold = False
        self.out = " "
        self.dirty = False

    @staticmethod
    def nullcolor(x, **kwargs):
        """
        A null color object which does not add any control codes to the output.

        Parameters
        ----------
        x : str
            The character being colored.
        **kwargs : dict
            Contains the bold flag.

        Returns
        -------
        str
            The value of x, unchanged.
        """
        return x

    def clear(self):
        """
        Resets the character cell to a space character.
        """
        self.out = "\033[38;5;220m \033[0m"
        self.dirty = False

    def output(self):
        """
        Calculates the output of the character cell if required, and returns its output.

        Returns
        -------
        str
            The output of this character cell, which is the calculated value of the out attribute.
        """
        if self.dirty:
            self.dirty = False
            out = self.value
            if self.color is not None:
                out = self.color(out, bold=self.bold)
            self.out = out
        return self.out


class Screen:
    """
    Defines a screen. Unless some serious magic is being done with multiple terminal windows, this does not need to be instantiated more than
    once.

    Attributes
    ----------
    ui : awsc.termui.ui.UI
        The UI instance which controls this screen.
    buf : list(awsc.termui.screen.Screen.Row)
        A list of character rows on the screen.
    """

    def __init__(self, ui):
        """
        Initializes a Screen object.

        Parameters
        ----------
        ui : awsc.termui.ui.UI
            The UI instance which controls this screen.
        """
        self.ui = ui
        self.buf = []
        self.clear()

    class Row:
        """
        Defines a row of characters on the screen.

        Attributes
        ----------
        screen : awsc.termui.screen.Screen
            The Screen object which controls this row.
        buf : list(awsc.termui.screen.Character)
            A list of character cells in the row.
        """

        def __init__(self, screen):
            """
            Initializes a Row object.

            Parameters
            ----------
            screen : awsc.termui.screen.Screen
                The Screen instance which controls this row.
            """
            self.screen = screen
            self.buf = []
            self.clear()

        def clear(self):
            """
            Clears the row, resetting each character to an empty cell. If the dimensions of the screen changed, this also recreates the buffer
            with fresh empty characters matching the width of the row.
            """
            if len(self.buf) == self.screen.ui.width:
                for character in self.buf:
                    character.clear()
            else:
                self.buf = [Character() for i in range(self.screen.ui.width)]

        def __getitem__(self, i):
            """
            Returns the i-th cell in the row.

            Parameters
            ----------
            i : int
                The index of the cell within the row to retrieve.

            Returns
            -------
            awsc.termui.screen.Character
                The i-th cell in the row.
            """
            return self.buf[i]

        def output(self):
            """
            Returns the output of all cells in the row, as a single string.

            Returns
            -------
            str
                The output for the entire row.
            """
            return "".join(x.output() for x in self.buf)

    def clear(self):
        """
        Clears the screen, resetting each row to an empty row. If the dimensions of the screen changed, this also recreates the buffer
        with fresh empty row matching the height of the screen.
        """
        if len(self.buf) == self.ui.height:
            for row in self.buf:
                row.clear()
        else:
            self.buf = [Screen.Row(self) for i in range(self.ui.height)]

    def __getitem__(self, i):
        """
        Returns the i-th row on the screen.

        Parameters
        ----------
        i : int
            The index of the row to retrieve.

        Returns
        -------
        awsc.termui.screen.Screen.Row
            The i-th row on the screen..
        """
        return self.buf[i]

    def output(self):
        """
        Returns the output of all rows on the screen, as a single string.

        Returns
        -------
        str
            The output for the entire screen.
        """
        print("\r\n".join(x.output() for x in self.buf), end="")



"""
Module for ELB v2 listener resources.
"""
from .arn import ARN
from .base_control import Describer, ResourceLister, ResourceRefByClass


class ListenerActionDescriber(Describer):
    """
    Describer control for listener rules.
    """

    prefix = "listener_action_browser"
    title = "Listener Rule"

    resource_type = "listener rule"
    main_provider = "elbv2"
    category = "ELB v2"
    subcategory = "Listener Rules"
    describe_method = "describe_rules"
    describe_kwarg_name = "RuleArns"
    describe_kwarg_is_list = True
    object_path = ".Rules[0]"
    default_entry_key = "arn"


def _listener_action_determine_condition(result, *args):
    """
    Column callback for determining conditions on listener rules.
    """
    if len(result["Conditions"]) > 1:
        return "<multiple>"
    if len(result["Conditions"]) == 1:
        cond = result["Conditions"][0]
        field = cond["Field"]
        if field == "host-header":
            if "Values" in cond and cond["Values"]:
                if isinstance(cond["Values"], str):
                    return f"Host: {cond['Values']}"
                src = cond["Values"]
            else:
                src = cond["HostHeaderConfig"]["Values"]
            return f"Host: {'|'.join(src)}"
        if field == "path":
            if "Values" in cond and cond["Values"]:
                if isinstance(cond["Values"], str):
                    return f"Path: {cond['Values']}"
                src = cond["Values"]
            else:
                src = cond["PathPatternConfig"]["Values"]
            return f"Path: {'|'.join(src)}"

        return field
    return "<always>"


def _listener_action_determine_action_type(result, *args):
    """
    Column callback for determining types on listener rules.
    """
    if len(result["Actions"]) > 0:
        act = result["Actions"][-1]
        return act["Type"]
    return "N/A"


def _listener_action_determine_target(result, *args):
    """
    Column callback for determining targetss on listener rules.
    """
    if len(result["Actions"]) > 0:
        act = result["Actions"][-1]
        if act["Type"] == "forward":
            return ",".join(
                [
                    ARN(i["TargetGroupArn"]).resource_id_first
                    for i in act["ForwardConfig"]["TargetGroups"]
                ]
            )
        if act["Type"] == "redirect":
            redirect = act["RedirectConfig"]
            proto = (
                "http(s)"
                if redirect["Protocol"] == "#{protocol}"
                else redirect["Protocol"].lower()
            )
            proto_ports = []
            if proto in ["http", "http(s)"]:
                proto_ports.append("80")
            if proto in ["https", "http(s)"]:
                proto_ports.append("443")
            port = f":{redirect['Port']}" if redirect["Port"] not in proto_ports else ""
            code = ("301" if redirect["StatusCode"] == "HTTP_301" else "302",)
            return f"{code} {proto}://{redirect['Host']}{port}{redirect['Path']}"
    return ""


class ListenerActionResourceLister(ResourceLister):
    """
    Lister control for listener rules.
    """

    prefix = "listener_action_list"
    title = "Listener Rules"

    resource_type = "listener rule"
    main_provider = "elbv2"
    category = "ELB v2"
    subcategory = "Listener Rules"
    list_method = "describe_rules"
    item_path = ".Rules"
    columns = {
        "priority": {
            "path": ".Priority",
            "size": 10,
            "weight": 0,
            "sort_weight": 0,
        },
        "condition": {
            "path": _listener_action_determine_condition,
            "size": 40,
            "weight": 1,
            "sort_weight": 1,
        },
        "action type": {
            "path": _listener_action_determine_action_type,
            "size": 20,
            "weight": 2,
            "sort_weight": 2,
        },
        "target": {
            "path": _listener_action_determine_target,
            "size": 120,
            "weight": 3,
            "sort_weight": 3,
        },
        "name": {"path": ".RuleArn", "hidden": True},
        "arn": {"path": ".RuleArn", "hidden": True},
    }
    primary_key = "arn"
    describe_command = ListenerActionDescriber.opener
    open_command = ResourceRefByClass("TargetGroupResourceLister")
    open_selection_arg = "rule_entry"

    def title_info(self):
        return self.listener["name"]

    def __init__(self, *args, listener, **kwargs):
        self.listener = listener
        self.list_kwargs = {"ListenerArn": self.listener["arn"]}
        super().__init__(*args, **kwargs)

    def open(self, *args):
        if self.selection is not None:
            if self.selection["action type"] == "forward":
                return super().open(*args)
            return self.describe(*args)
        return None


class ListenerDescriber(Describer):
    """
    Describer control for ELB v2 listeners.
    """

    prefix = "listener_browser"
    title = "Listener"

    resource_type = "listener"
    main_provider = "elbv2"
    category = "ELB v2"
    subcategory = "Listener"
    describe_method = "describe_listeners"
    describe_kwarg_name = "ListenerArns"
    describe_kwarg_is_list = True
    object_path = ".Listeners[0]"


def _listener_determine_ssl_policy(self, result, *args):
    """
    Column callback which returns SSL Policy when it can exist.
    """
    if result["Protocol"] == "HTTPS":
        return result["SslPolicy"]
    return ""


class ListenerResourceLister(ResourceLister):
    """
    Lister control for ELB v2 listeners.
    """

    prefix = "listener_list"
    title = "Listeners"

    resource_type = "listener"
    main_provider = "elbv2"
    category = "ELB v2"
    subcategory = "Listeners"
    list_method = "describe_listeners"
    item_path = ".Listeners"
    columns = {
        "protocol": {
            "path": ".Protocol",
            "size": 10,
            "weight": 0,
            "sort_weight": 1,
        },
        "port": {"path": ".Port", "size": 10, "weight": 1, "sort_weight": 0},
        "instance type": {
            "path": _listener_determine_ssl_policy,
            "size": 20,
            "weight": 2,
        },
        "name": {"path": ".ListenerArn", "hidden": True},
        "arn": {"path": ".ListenerArn", "hidden": True},
    }
    primary_key = "arn"
    describe_command = ListenerDescriber.opener
    open_command = ListenerActionResourceLister.opener
    open_selection_arg = "listener"

    def title_info(self):
        return self.load_balancer["name"]

    def __init__(self, *args, lb, **kwargs):
        self.load_balancer = lb
        self.list_kwargs = {"LoadBalancerArn": self.load_balancer["arn"]}
        super().__init__(*args, **kwargs)



"""
Module for region-related resources.
"""
from .base_control import OpenableListControl
from .common import Common
from .termui.list_control import ListEntry


class RegionList(OpenableListControl):
    """
    Lister control for available AWS regions.
    """

    prefix = "region_list"
    title = "Regions"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_column("usage frequency", 12)
        self.add_column("default", 8)
        regions = sorted(Common.Session.service_provider.list_regions())
        for region in regions:
            self.add_entry(
                ListEntry(
                    region,
                    **{
                        "usage frequency": 0,
                        "default": "✓"
                        if region == Common.Configuration["default_region"]
                        else " ",
                    }
                )
            )
        try:
            self.selected = regions.index(Common.Configuration["default_region"])
        except ValueError:
            self.selected = 0

    @OpenableListControl.Autohotkey("d", "Set as default", True)
    def set_default_region(self, _):
        """
        Hotkey callback for setting the default region.
        """
        Common.Configuration["default_region"] = self.selection.name
        Common.Configuration.write_config()
        for entry in self.entries:
            if entry.name != Common.Configuration["default_region"]:
                entry["default"] = " "
            else:
                entry["default"] = "✓"

    @OpenableListControl.Autohotkey("KEY_ENTER", "Select", True)
    def select_region(self, _):
        """
        Hotkey callback for setting the active region.
        """
        Common.Session.region = self.selection.name



"""
Module for RDS-related resources.
"""
import subprocess

from .base_control import Describer, ResourceLister, tag_finder_generator
from .common import Common, SessionAwareDialog
from .termui.control import Border
from .termui.dialog import DialogFieldText


class RDSDescriber(Describer):
    """
    Describer control for RDS databases.
    """

    prefix = "rds_browser"
    title = "RDS Instance"

    resource_type = "database instance"
    main_provider = "rds"
    category = "RDS"
    subcategory = "Database Instance"
    describe_method = "describe_route_tables"
    describe_kwarg_name = "DBInstanceIdentifier"
    object_path = ".DBInstances[0]"
    default_entry_key = "instance id"


class RDSResourceLister(ResourceLister):
    """
    Lister control for RDS databases.
    """

    prefix = "rds_list"
    title = "RDS Instances"
    command_palette = ["rds"]

    resource_type = "database instance"
    main_provider = "rds"
    category = "RDS"
    subcategory = "Database Instances"
    list_method = "describe_db_instances"
    item_path = ".DBInstances"
    columns = {
        "instance id": {
            "path": ".DBInstanceIdentifier",
            "size": 11,
            "weight": 0,
            "sort_weight": 1,
        },
        "host": {
            "path": ".Endpoint.Address",
            "size": 45,
            "weight": 1,
            "sort_weight": 0,
        },
        "engine": {
            "path": ".Engine",
            "size": 10,
            "weight": 2,
        },
        "type": {
            "path": ".DBInstanceClass",
            "size": 10,
            "weight": 3,
        },
        "vpc": {
            "path": ".DBSubnetGroup.VpcId",
            "size": 15,
            "weight": 4,
        },
        "public_access": {"path": ".PubliclyAccessible", "hidden": True},
        "db_name": {"path": ".PubliclyAccessible", "hidden": True},
        "name": {
            "path": tag_finder_generator("Name", taglist_key="TagList"),
            "hidden": True,
        },
    }
    describe_command = RDSDescriber.opener
    primary_key = "instance id"

    @ResourceLister.Autohotkey("s", "Open command line", True)
    def db_client(self, _):
        """
        Attempts to open a command line session to connect to the RDS database.
        """
        if self.selection["public_access"] != "True":
            Common.Session.set_message(
                "No public IP associated with instance",
                Common.color("message_error"),
            )
        else:
            RDSClientDialog.opener(instance_entry=self.selection, caller=self)


class RDSClientDialog(SessionAwareDialog):
    """
    Dialog for opening a command line session to connect to the RDS database.

    Attributes
    ----------
    instance_id : str
        The ID of the RDS instance we're connecting to.
    db_name : str
        The name of the default database.
    ip : str
        The hostname of the database.
    username_textfield : awsc.termui.dialog.DialogFieldText
        The username for logging into the database.
    password_textfield : awsc.termui.dialog.DialogFieldText
        The password for logging into the database.
    database_textfield : str
        Optionally, connect to a different database within the instance.
    caller : awsc.termui.control.Control
        The parent control controlling this dialog.
    """

    line_size = 15

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        *args,
        instance_entry=None,
        caller=None,
        **kwargs,
    ):
        kwargs["border"] = Border(
            Common.border("rds_client_modal", "default"),
            Common.color("rds_client_modal_dialog_border", "modal_dialog_border"),
            "SSH to instance",
            Common.color(
                "rds_client_modal_dialog_border_title", "modal_dialog_border_title"
            ),
            instance_entry["instance id"],
            Common.color(
                "rds_client_modal_dialog_border_title_info",
                "modal_dialog_border_title_info",
            ),
        )
        super().__init__(parent, alignment, dimensions, *args, caller=caller, **kwargs)
        self.instance_id = instance_entry["instance id"]
        self.db_name = instance_entry["db_name"]
        self.ip = instance_entry["host"]
        self.engine = instance_entry["engine"]
        self.set_title_label("Connect to RDS instance")
        self.username_textfield = DialogFieldText(
            "Username",
            text="",
            color=Common.color(
                "rds_client_modal_dialog_textfield", "modal_dialog_textfield"
            ),
            selected_color=Common.color(
                "rds_client_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
            label_color=Common.color(
                "rds_client_modal_dialog_textfield_label",
                "modal_dialog_textfield_label",
            ),
            label_min=16,
        )
        self.password_textfield = DialogFieldText(
            "Password",
            text="",
            color=Common.color(
                "rds_client_modal_dialog_textfield", "modal_dialog_textfield"
            ),
            selected_color=Common.color(
                "rds_client_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
            label_color=Common.color(
                "rds_client_modal_dialog_textfield_label",
                "modal_dialog_textfield_label",
            ),
            label_min=16,
            password=True,
        )
        self.database_textfield = DialogFieldText(
            "Database",
            text=self.db_name,
            color=Common.color(
                "rds_client_modal_dialog_textfield", "modal_dialog_textfield"
            ),
            selected_color=Common.color(
                "rds_client_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
            label_color=Common.color(
                "rds_client_modal_dialog_textfield_label",
                "modal_dialog_textfield_label",
            ),
            label_min=16,
        )
        self.add_field(self.username_textfield)
        self.add_field(self.password_textfield)
        self.add_field(self.database_textfield)
        # self.highlighted = 1 if def_text != "" else 0
        self.caller = caller

    def accept_and_close(self):
        if self.engine in ["aurora", "aurora-mysql", "mariadb", "mysql"]:
            dollar_zero = "mysql"
            cmd = f"mysql -h {self.ip} -D {self.database_textfield.text} -u {self.username_textfield.text} --password={self.password_textfield.text}"
        elif self.engine in ["aurora-postgresql", "postgres"]:
            dollar_zero = "psql"
            cmd = f"psql postgres://{self.username_textfield}:{self.password_textfield}@{self.ip}/{self.database_textfield}"
        else:
            Common.Session.set_message(
                f"Unsupported engine: {self.engine}",
                Common.color("message_info"),
            )
            self.close()
            return
        exit_code = Common.Session.ui.unraw(subprocess.run, ["bash", "-c", cmd])
        Common.Session.set_message(
            f"{dollar_zero} exited with code {exit_code.returncode}",
            Common.color("message_info"),
        )
        super().accept_and_close()



"""
EC2 AMI resource controls.
"""
from .base_control import Describer, ResourceLister


class AMIDescriber(Describer):
    """
    Describer control for AMI resources.
    """

    prefix = "ami_browser"
    title = "Amazon Machine Image"

    resource_type = "AMI"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "AMI"
    describe_method = "describe_images"
    describe_kwarg_name = "ImageIds"
    describe_kwarg_is_list = True
    object_path = ".Images[0]"
    default_entry_key = "id"


class AMIResourceLister(ResourceLister):
    """
    List control for AMI resources.

    Attributes
    ----------
    ec2 : awsc.termui.list_control.ListEntry
        List AMIs in the context of this EC2 instance.
    """

    prefix = "ami_list"
    title = "Amazon Machine Images"
    command_palette = ["ami", "image"]

    resource_type = "AMI"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "AMI"
    list_method = "describe_images"
    list_kwargs = {"Owners": ["self"]}
    item_path = ".Images"
    primary_key = "id"
    columns = {
        "id": {
            "path": ".ImageId",
            "size": 15,
            "weight": 0,
            "sort_weight": 0,
        },
        "name": {"path": ".Name", "size": 64, "weight": 1},
        "arch": {"path": ".Architecture", "size": 8, "weight": 2},
        "platform": {"path": ".PlatformDetails", "size": 10, "weight": 3},
        "type": {"path": ".ImageType", "size": 10, "weight": 4},
        "owner": {"path": ".ImageOwnerAlias", "size": 15, "weight": 5},
        "state": {"path": ".State", "size": 10, "weight": 6},
        "virt": {"path": ".VirtualizationType", "size": 10, "weight": 7},
    }

    describe_selection_arg = "entry"
    describe_command = AMIDescriber.opener

    def title_info(self):
        return None if self.ec2 is None else f"Instance: {self.ec2['instance id']}"

    def __init__(self, *args, ec2=None, **kwargs):
        self.ec2 = ec2
        if ec2 is not None:
            self.list_kwargs["ImageIds"] = [kwargs["ec2"]["image"]]
        super().__init__(*args, **kwargs)



"""
Module for the configuration parser object.
"""
import sys
from pathlib import Path

import yaml

from .scheme import Scheme
from .storage import Keystore

LAST_CONFIG_VERSION = 19


class Config:
    """
    Configuration parser and holder.

    Attributes
    ----------
    version_updaters : dict
        A numbered dict of update functions for incrementally upgrading an outdated configuration file.
    path : pathlib.Path
        The configuration parent path, where all configuration data is stored.
    config_path : pathlib.Path
        The path to the configuration yaml file.
    keystore : awsc.config.storage.Keystore
        The key storage holder instance.
    scheme : awsc.config.scheme.Scheme
        The color scheme holder instance.
    config : dict
        The loaded configuration.
    """

    def __init__(self, path=None):
        """
        Initializes a Config object.

        Parameters
        ----------
        path : str
            The configuration parent path, if not default. Defaults to ~/.config/awsc.
        """
        print("Initializing AWSC configuration...", file=sys.stderr)
        self.version_updaters = {
            1: self._update_1,
            2: self._update_2,
            3: self._update_3,
            4: self._update_4,
            5: self._update_5,
            6: self._update_6,
            7: self._update_7,
            8: self._update_8,
            9: self._update_9,
            10: self._update_10,
            11: self._update_11,
            12: self._update_12,
            13: self._update_13,
            14: self._update_14,
            15: self._update_15,
            16: self._update_16,
            17: self._update_17,
            18: self._update_18,
            19: self._update_19,
        }
        if path is None:
            path = Path.home() / ".config" / "awsc"
        else:
            path = Path(path)
        if not path.exists():
            path.mkdir(parents=True, exist_ok=True)
        self.path = path
        self.keystore = Keystore(self)
        self.scheme = Scheme(self)

        self.config_path = self.path / "config.yaml"

        if not self.config_path.exists():
            self.create_default_config()
        else:
            self.parse_config()

    def initialize(self):
        """
        Late initializer for the Config object. Should be called after everything has been instantiated
        but before anything is being used.
        """
        self.keystore.unlock()
        self.update_version()

    def _update_1(self):
        """
        Version 1 configuration update.

        Do not call.
        """
        self.config["default_region"] = "us-east-1"

    def _update_2(self):
        """
        Version 2 configuration update.

        Do not call.
        """
        self.scheme.backup_and_reset()

    def _update_3(self):
        """
        Version 3 configuration update.

        Do not call.
        """
        self.scheme.backup_and_reset()

    def _update_4(self):
        """
        Version 4 configuration update.

        Do not call.
        """
        self.scheme.backup_and_reset()
        self.config["usage_statistics"] = {
            "regions_by_context": {},
            "resources_by_context": {},
        }

    def _update_5(self):
        """
        Version 5 configuration update.

        Do not call.
        """
        self.scheme.backup_and_reset()

    def _update_6(self):
        """
        Version 6 configuration update.

        Do not call.
        """
        self.config["default_ssh_key"] = "id_rsa"

    def _update_7(self):
        """
        Version 7 configuration update.

        Do not call.
        """
        self.scheme.backup_and_reset()
        self.config["default_ssh_usernames"] = {}

    def _update_8(self):
        """
        Version 8 configuration update.

        Do not call.
        """
        self.config["editor_command"] = "nano {0}"

    def _update_9(self):
        """
        Version 9 configuration update.

        Do not call.
        """
        self.config["keypair_associations"] = {}

    def _update_10(self):
        """
        Version 10 configuration update.

        Do not call.
        """
        self.scheme.backup_and_reset()

    def _update_11(self):
        """
        Version 11 configuration update.

        Do not call.
        """
        self.scheme.backup_and_reset()

    def _update_12(self):
        """
        Version 12 configuration update.

        Do not call.
        """
        self.config["log_retention"] = {
            "max_lines": -1,
            "max_age": 2419200,
        }

    def _update_13(self):
        """
        Version 13 configuration update.

        Do not call.
        """
        self.config["default_dashboard_layout"] = [
            ["Blank", "Blank"],
            ["Blank", "Blank"],
        ]
        self.config["dashboard_layouts"] = {}
        self.scheme.backup_and_reset()

    def _update_14(self):
        """
        Version 14 configuration update.

        Do not call.
        """
        self.config["contexts"]["localstack"] = {
            "endpoint_url": "https://localhost.localstack.cloud"
        }

    def _update_15(self):
        """
        Version 15 configuration update.

        Do not call.
        """
        self.config["contexts"]["localstack"]["account_id"] = "localhost"

    def _update_16(self):
        """
        Version 16 configuration update.

        Do not call.
        """
        self.config["contexts"]["localstack"][
            "endpoint_url"
        ] = "https://localhost.localstack.cloud:4566"

    def _update_17(self):
        """
        Version 17 configuration update.

        Do not call.
        """
        for context in self.config["contexts"].keys():
            self.config["contexts"][context]["mfa_device"] = ""

    def _update_18(self):
        """
        Version 18 configuration update.

        Do not call.
        """
        error_path = Path.home() / ".config" / "awsc" / "error.log"
        self.config["error_log"] = str(error_path)

    def _update_19(self):
        """
        Version 19 configuration update.

        Do not call.
        """
        for context in self.config["contexts"].keys():
            self.config["contexts"][context]["role"] = ""

    def update_version(self):
        """
        Main configuration update sequence. Always called by initialize(), not required to call separately.
        """
        if "version" not in self.config:
            version = 0
        else:
            version = self.config["version"]
        while version < LAST_CONFIG_VERSION:
            print(f"Performing config update to version {version + 1}")
            self.version_updaters[version + 1]()
            version += 1
        self.config["version"] = version
        self.write_config()

    def create_default_config(self):
        """
        Generates first time default configuration. Should match the expected format of the latest update version.

        Writes the generated configuration to the configuration file in config_path.
        """
        print("Creating first time configuration...")
        error_path = Path.home() / ".config" / "awsc" / "error.log"
        self.config = {
            "version": LAST_CONFIG_VERSION,
            "contexts": {
                "localstack": {
                    "endpoint_url": "https://localhost.localstack.cloud:4566",
                    "account_id": "localhost",
                }
            },
            "default_context": "",
            "default_region": "us-east-1",
            "default_ssh_key": "id_rsa",
            "default_ssh_usernames": {},
            "keypair_associations": {},
            "editor_command": "nano {0}",
            "log_retention": {
                "max_lines": -1,
                "max_age": 2419200,
            },
            "usage_statistics": {
                "regions": {},
                "resources": {},
                "keys": {},
            },
            "default_dashboard_layout": [["Blank", "Blank"], ["Blank", "Blank"]],
            "dashboard_layouts": {},
            "error_log": str(error_path),
        }

        self.write_config()

    def write_config(self):
        """
        Immediately writes the configuration to the configuration file in config_path.
        """
        with self.config_path.open("w", encoding="utf-8") as file:
            file.write(yaml.dump(self.config))

    def __contains__(self, item):
        return item in self.config

    def __getitem__(self, item):
        """
        Retrieve a configuration value.

        Parameters
        ----------
        item : str
            The configuration key.

        Returns
        -------
        object
            The configuration value.
        """
        return self.config[item]

    def __setitem__(self, item, value):
        """
        Set a configuration value.

        Parameters
        ----------
        item : str
            The configuration key.
        value : object
            The configuration value.
        """
        self.config[item] = value

    def parse_config(self):
        """
        Parses the configuration file specified in config_path and replaces the currently loaded config with its contents.
        """
        with self.config_path.open("r", encoding="utf-8") as file:
            self.config = yaml.safe_load(file.read())

    def add_or_edit_context(self, name, acctid, access, secret, mfa_device=""):
        """
        Upserts a context into the configuration.

        Parameters
        ----------
        name : str
            The name of the context. If it already exists, it will be overwritten.
        acctid : str
            The account number of the context, usually acquired through Common.Session.whoami()
        access : str
            The access key which belongs to the account.
        secret : str
            The secret key which belongs to the account.
        """
        self.config["contexts"][name] = {
            "account_id": acctid,
            "mfa_device": mfa_device,
            "role": "",
        }
        if self.config["default_context"] == "":
            self.config["default_context"] = name
        self.keystore.set_key(name, access, secret)
        self.write_config()

    def add_or_edit_role_context(self, name, acctid, source, role, mfa_device=""):
        """
        Upserts a context into the configuration.

        Parameters
        ----------
        name : str
            The name of the context. If it already exists, it will be overwritten.
        acctid : str
            The account number of the context, usually acquired through Common.Session.whoami()
        access : str
            The access key which belongs to the account.
        secret : str
            The secret key which belongs to the account.
        """
        self.config["contexts"][name] = {
            "account_id": acctid,
            "mfa_device": mfa_device,
            "role": role,
        }
        self.keystore.set_ref(name, source)
        self.write_config()

    def delete_context(self, name):
        """
        Deletes a context from the configuration.

        Parameters
        ----------
        name : str
            The name of the context.
        """
        del self.config["contexts"][name]
        if self.config["default_context"] == name:
            if len(self.config["contexts"]):
                self.config["default_context"] = list(self.config["contexts"].keys())[0]
        self.keystore.delete_key(name)
        self.write_config()



"""
Module for IAM role resources.
"""

from .base_control import (
    Describer,
    ListResourceDocumentCreator,
    ResourceLister,
    SelectionAttribute,
    SelectorTemplateField,
    TemplateDict,
)
from .resource_iam import InlinePolicyAttacher
from .termui.ui import ControlCodes


class RoleDescriber(Describer):
    """
    Describer control for IAM Role resources.
    """

    prefix = "role_browser"
    title = "Role"

    resource_type = "role"
    main_provider = "iam"
    category = "IAM"
    subcategory = "Role"
    describe_method = "get_role"
    describe_kwarg_name = "RoleName"
    object_path = ".Role"


@ResourceLister.Autocommand(
    "InstanceProfileLister", "u", "View roles", "instanceprofile"
)
class RoleLister(ResourceLister):
    """
    Lister control for IAM Role resources.
    """

    prefix = "role_list"
    title = "Roles"
    command_palette = ["role", "roles"]

    resource_type = "role"
    main_provider = "iam"
    category = "IAM"
    subcategory = "Role"
    columns = {
        "name": {
            "path": ".RoleName",
            "size": 25,
            "weight": 0,
            "sort_weight": 0,
        },
        "id": {"path": ".RoleId", "size": 15, "weight": 1},
        "path": {"path": ".Path", "size": 40, "weight": 2},
        "arn": {"path": ".Arn", "hidden": True},
    }
    describe_command = RoleDescriber.opener
    open_command = "m"

    def title_info(self):
        return (
            None
            if self.instanceprofile is None
            else f"InstanceProfile: {self.instanceprofile['name']}"
        )

    def create_role(self, _):
        """
        Hotkey callback for creating a new role.
        """
        creator = ListResourceDocumentCreator(
            "iam",
            "create_role",
            None,
            as_json=["Tags"],
            initial_document={
                "RoleName": "",
                "Path": "/",
                "PermissionsBoundary": "",
                "MaxSessionDuration": 3600,
                "AssumeRolePolicyDocument": {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Sid": "",
                            "Effect": "Allow",
                            "Action": ["sts:AssumeRole"],
                            "Resource": "*",
                            "Condition": {},
                            "Principal": {},
                        }
                    ],
                },
                "Description": "",
                "Tags": [{"Key": "", "Value": ""}],
            },
        )
        creator.edit()
        self.refresh_data()

    @ResourceLister.Autohotkey("a", "Attach policy", True)
    def attach_policy(self, _):
        """
        Hotkey callback for attaching a policy to a role.
        """
        self.selector_template(
            "attach_role_policy",
            TemplateDict(
                {
                    "RoleName": SelectionAttribute("name"),
                    "PolicyArn": SelectorTemplateField(),
                }
            ),
            "Attach policy to role",
            "attach",
            "PolicyLister",
            resource_type="policy",
            success_template="Attaching policy {PolicyArn} to role {RoleName}",
        )(self.selection)

    def add_to_instance_profile(self, _):
        """
        Hotkey callback for adding the role to an instance profile.
        """
        self.selector_template(
            "add_role_to_instance_profile",
            TemplateDict(
                {
                    "RoleName": SelectionAttribute("name"),
                    "InstanceProfileName": SelectorTemplateField(),
                }
            ),
            "Add role to instance profile",
            "add",
            "InstanceProfileLister",
            resource_type="group",
            success_template="Adding role {RoleName} to instance profile {InstanceProfileName}",
        )(self.selection)

    def __init__(self, *args, instanceprofile=None, **kwargs):
        self.instanceprofile = instanceprofile
        self.list_method = (
            "list_roles" if self.instanceprofile is None else "get_instance_profile"
        )
        self.list_kwargs = (
            {}
            if self.instanceprofile is None
            else {"InstanceProfileName": self.instanceprofile["name"]}
        )
        self.item_path = (
            ".Roles" if self.instanceprofile is None else ".InstanceProfile.Roles"
        )

        super().__init__(*args, **kwargs)
        self.inliner = InlinePolicyAttacher(self, "role")
        if self.instanceprofile is None:
            self.add_hotkey("n", self.create_role, "Create Role")
            self.add_hotkey(
                "j", self.add_to_instance_profile, "Add to Instance Profile", True
            )
            self.confirm_template(
                "delete_role",
                TemplateDict(
                    {
                        "RoleName": SelectionAttribute("name"),
                    }
                ),
                hotkey=ControlCodes.D,
                hotkey_tooltip="Delete Role",
            )
        else:
            self.confirm_template(
                "remove_role_from_instance_profile",
                TemplateDict(
                    {
                        "RoleName": SelectionAttribute("name"),
                        "InstanceProfileName": self.instanceprofile["name"],
                    }
                ),
                success_template="Removing role {RoleName} from instance profile {InstanceProfileName}",
                action_name="Detach",
                from_what="instance profile",
                from_what_name=self.instanceprofile["name"],
                hotkey=ControlCodes.D,
                hotkey_tooltip="Detach from Instance Profile",
            )



"""
Module for EC2 instance class resources.
"""
from .base_control import Describer, ResourceLister


class InstanceClassDescriber(Describer):
    """
    Describer control for EC2 instance classes.
    """

    prefix = "instance_class_browser"
    title = "Instance Class"

    resource_type = "instance class"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "Instance Class"
    describe_method = "describe_instance_types"
    describe_kwarg_name = "InstanceTypes"
    describe_kwarg_is_list = True
    object_path = ".InstanceTypes[0]"


class InstanceClassResourceLister(ResourceLister):
    """
    List control for EC2 instance classes.
    """

    prefix = "instance_class_list"
    title = "Instance Classes"
    command_palette = ["it", "instancetype", "instanceclass"]

    resource_type = "instance class"
    main_provider = "ec2"
    category = "EC2"
    subcategory = "Instance Class"
    list_method = "describe_instance_types"
    item_path = ".InstanceTypes"
    list_kwargs = {"MaxResults": 20}
    columns = {
        "name": {
            "path": ".InstanceType",
            "size": 20,
            "weight": 0,
            "sort_weight": 0,
        },
        "cpus": {"path": ".VCpuInfo.DefaultVCpus", "size": 4, "weight": 1},
        "memory": {
            "path": lambda x: f"{x['MemoryInfo']['SizeInMiB']} MiB",
            "size": 10,
            "weight": 2,
        },
        "ebs optimization": {
            "path": ".EbsInfo.EbsOptimizedSupport",
            "size": 15,
            "weight": 3,
        },
        "network": {"path": ".NetworkInfo.NetworkPerformance", "size": 15, "weight": 4},
    }
    describe_command = InstanceClassDescriber.opener



"""
Module for common storage and functions.
"""
import configparser
import json
import sys
import traceback
from datetime import datetime, timezone
from pathlib import Path
from typing import Callable, Set

import yaml
from botocore import exceptions

from .config.config import Config
from .session import Session
from .termui.alignment import CenterAnchor, Dimension, TopLeftAnchor
from .termui.bar_graph import BarGraph
from .termui.color import Color, Palette8Bit
from .termui.control import Border, BorderStyle
from .termui.dialog import DialogControl, DialogFieldLabel

DefaultAnchor = TopLeftAnchor(0, 11)
DefaultDimension = Dimension("100%", "100%-14")


def default_args():
    """
    Generates the default args for opening a new Block at the top level. Convenience function.

    Returns
    -------
    list
        Pass this as the first *argument of a Block subclass constructor.
    """
    return [Common.Session.ui.top_block, DefaultAnchor, DefaultDimension]


def default_kwargs(prefix):
    """
    Generates the default kwargs for opening a new Control at the top level. Convenience function.

    Arguments
    ---------
    prefix : str
        The prefix for the color scheme of this Control.

    Returns
    -------
    dict
        Pass this as the **keyword arguments of a Control subclass constructor.
    """
    return {
        "weight": 0,
        "color": Common.color(f"{prefix}_generic", "generic"),
        "selection_color": Common.color(f"{prefix}_selection", "selection"),
        "filtered_color": Common.color(f"{prefix}_filtered", "selection"),
        "title_color": Common.color(f"{prefix}_heading", "column_title"),
        "update_color": Common.color(f"{prefix}_updated", "highlight"),
        "update_selection_color": Common.color(
            f"{prefix}_updated", "highlight_selection"
        ),
    }


def datetime_hack(x):
    """
    JSON dumper hack, as the base implementation of datetime doesn't understand how to convert it to JSON.

    Parameters
    ----------
    x : object
        The object the JSON dumper could not handle.

    Returns
    -------
    str
        The ISO representation of the datetime object.

    Raises
    ------
    TypeError
        If x is not a datetime.datetime object.
    """
    if isinstance(x, datetime):
        return x.isoformat()
    raise TypeError("Unknown type")


# TODO: Should be in base_control.
class BaseChart(BarGraph):
    """
    Base class for chart controls.
    """

    prefix = "CHANGEME"
    title = "CHANGEME"

    @classmethod
    def opener(cls, *args, **kwargs):
        """
        Session-aware initializer for this class. Creates a BaseChart object with the default alignment settings for the awsc layout.

        Returns
        -------
        list(awsc.common.BaseChart)
            A new instance of this class.
        """
        ret = cls(
            Common.Session.ui.top_block,
            DefaultAnchor,
            DefaultDimension,
            *args,
            weight=0,
            color=Common.color("generic"),
            **kwargs,
        )
        ret.border = Common.default_border(cls.prefix, cls.title, ret.title_info())
        return [ret]

    def title_info(self):
        """
        Title information callback.

        Returns
        -------
        str
            The extra info to display in the title.
        """
        return None


# TODO: Should be in base_control or separate dialog module.
class SessionAwareDialog(DialogControl):
    """
    Base class for dialogs.
    """

    line_size = 10

    @classmethod
    def opener(cls, *args, caller, **kwargs):
        """
        Session-aware initializer for this class. Creates a SessionAwareDialog object with the default alignment settings for the awsc layout.

        Returns
        -------
        awsc.common.SessionAwareDialog
            A new instance of this class.
        """
        return cls(
            caller.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", f"{cls.line_size}"),
            *args,
            caller=caller,
            weight=-500,
            **kwargs,
        )

    def __init__(self, *args, caller, **kwargs):
        caller.dialog_mode = True
        self.caller = caller
        kwargs["ok_action"] = self.accept_and_close
        kwargs["cancel_action"] = self.close
        super().__init__(*args, **kwargs)
        Common.Session.extend_frame(self)
        self.title_label = DialogFieldLabel("TITLE")
        self.add_field(self.title_label)
        self.add_field(DialogFieldLabel(""))

        self.error_label = DialogFieldLabel(
            "", default_color=Common.color("modal_dialog_error")
        )
        self.add_field(self.error_label)
        self.add_field(DialogFieldLabel(""))

    def set_title_label(self, text):
        """
        Sets the text on the title label.

        Parameters
        ----------
        text : str
            The new title.
        """
        self.title_label.text = text

    def input(self, key):
        if key.is_sequence and key.name == "KEY_ESCAPE":
            self.close()
            return True
        return super().input(key)

    def accept_and_close(self):
        """
        Callback for the affirmative action of the dialog. This should handle the data entered in the dialog, then execute the negative action.
        """
        self.close()

    def close(self):
        """
        Callback for the negative action of the dialog. This should _not_ handle the data entered in the dialog.
        """
        self.caller.dialog_mode = False
        self.parent.remove_block(self)
        Common.Session.remove_from_frame(self)


class Common:
    """
    Namespace class for common features used by awsc.

    Attributes
    ----------
    Configuration : awsc.config.config.Config
        The AWSC configuration.
    Session : awsc.session.Session
        The session manager object.
    _logholder : awsc.common.LogHolder
        The log stash object.
    initialized : bool
        Whether Common was initialized.
    init_hooks : set(callable)
        A set of hooks to execute after initialization.
    """

    Configuration: Config = Config()
    Session = None
    _logholder = None
    initialized = False
    init_hooks: Set[Callable[[], None]] = set()

    @classmethod
    def run_on_init(cls, hook):
        """
        Register a hook to run on initialization.

        Runs the hook immediately if already initialized.

        Parameters
        ----------
        hook : callable
            The hook to register.
        """
        if cls.initialized:
            hook()
        else:
            cls.init_hooks.add(hook)

    @classmethod
    def initialize(cls):
        """
        Initialize Common.
        """
        if cls.initialized:
            return
        cls.Configuration.initialize()
        cls.Session = Session(
            cls.Configuration,
            cls.color("info_display_title"),
            cls.color("info_display_value"),
            cls,
        )
        cls._logholder = LogHolder()
        cls.initialized = True

    @classmethod
    def post_initialize(cls):
        """
        Post-initialization actions for Common. Should be invoked from the outside, after initializing.
        """
        for hook in cls.init_hooks:
            hook()
        cls.load_dot_aws()

    @classmethod
    def load_dot_aws(cls):
        """
        Parse and load the contents of ~/.awsc/credentials. Adds contexts registered in that file as awsc contexts.
        """
        aws_creds = Path.home() / ".aws" / "credentials"
        print("Loading ~/.aws/credentials", file=sys.stderr)
        try:
            with aws_creds.open("r", encoding="utf-8") as file:
                creds = file.read()
        except OSError as error:
            print(
                f"Failed to open ~/.aws/credentials: {str(error)}",
                file=sys.stderr,
            )
            return
        parser = configparser.ConfigParser(default_section="__default")
        parser.read_string(creds)
        for section in parser.sections():
            if "aws_security_token" in parser[section]:
                print(
                    f"Skipping credentials {section} as they are security token authenticated"
                )
                continue
            if "aws_access_key_id" not in parser[section]:
                print(
                    f"aws_access_key_id missing for credential {section}, skipping",
                    file=sys.stderr,
                )
                continue
            if "aws_secret_access_key" not in parser[section]:
                print(
                    f"aws_secret_access_key missing for credential {section}, skipping",
                    file=sys.stderr,
                )
                continue
            access = parser[section]["aws_access_key_id"]
            secret = parser[section]["aws_secret_access_key"]
            api_keypair = {"access": access, "secret": secret}
            try:
                whoami = cls.Session.service_provider.whoami(keys=api_keypair)
            except exceptions.ClientError as error:
                cls.clienterror(
                    error,
                    "Verify Credentials",
                    "Bootstrap",
                    subcategory="Credentials Import",
                    resource=section,
                    set_message=False,
                    api_provider="sts",
                    api_method="get_caller_identity",
                    api_keypair=api_keypair,
                    api_args={},
                    credentials_section=section,
                )
                continue
            mfa_device = (
                ""
                if "aws_mfa_device" not in parser[section]
                else parser[section]["aws_mfa_device"]
            )
            cls.Configuration.add_or_edit_context(
                section, whoami["Account"], access, secret, mfa_device=mfa_device
            )
            print(
                f"Added {section} context from aws credentials file",
                file=sys.stderr,
            )

    @staticmethod
    def color(name, fallback=None):
        """
        Fetch a scheme color by name, with an optional fallback color if the main requested color is not present.
        """
        if Common.Configuration is None:
            raise ValueError("Configuration is not initialized.")
        if name not in Common.Configuration.scheme["colors"]:
            if fallback is None:
                raise KeyError(f'Undefined color "{name}"')
            return Common.color(fallback)
        return Color(
            Palette8Bit(),
            Common.Configuration.scheme["colors"][name]["foreground"],
            background=Common.Configuration.scheme["colors"][name]["background"],
        )

    @staticmethod
    def border(name, fallback=None):
        """
        Fetch a scheme border style by name, with an optional fallback border style if the main requested style is not present.
        """
        if Common.Configuration is None:
            raise ValueError("Configuration is not initialized.")
        if name not in Common.Configuration.scheme["borders"]:
            if fallback is None:
                raise KeyError(f'Undefined border "{name}"')
            return Common.border(fallback)
        border = Common.Configuration.scheme["borders"][name]
        return BorderStyle(
            [
                border["horizontal"],
                border["vertical"],
                border["TL"],
                border["TR"],
                border["BL"],
                border["BR"],
            ]
        )

    @staticmethod
    def main():
        """
        Main loop method. Passes main loop control to termui.
        """
        Common.Session.ui.main()

    @staticmethod
    def confdir():
        """
        Configuration directory retriever.

        Returns
        -------
        pathlib.Path
            The path of the configuration directory.
        """
        return Path.home() / ".config" / "awsc"

    @staticmethod
    def textlog(message):
        """
        Logs text to stderr, for instances where internal logging doesn't cut it.

        Parameters
        ----------
        message : str
            The message to log.
        """
        print(message, file=sys.stderr)

    @classmethod
    def log(
        cls,
        message,
        summary,
        category,
        message_type,
        subcategory=None,
        resource=None,
        set_message=True,
        **kwargs,
    ):
        """
        Logs a message to the log stash and saves it to disk.

        Parameters
        ----------
        message : str
            The full message to log.
        summary : str
            A short summary of the event being logged.
        category : str
            The category of the logged event.
        message_type : str
            The message type. Severity, but simpler. Expects one of "success", "info" or "error", but more may be defined as long as the scheme
            defines a color for it.
        subcategory : str
            The subcategory of the logged event.
        resource : str
            The resource identifier which triggered the event.
        set_message : bool
            Whether to also flash a message on the awsc message display about this event. The flashed message is the message.
        kwargs : dict
            Any additional keyword arguments will be logged as context for the event.
        """
        cls._logholder.log(
            message,
            summary,
            category,
            message_type,
            subcategory=subcategory,
            resource=resource,
            set_message=set_message,
            **kwargs,
        )

    @classmethod
    def error(
        cls,
        message,
        summary,
        category,
        subcategory=None,
        resource=None,
        set_message=True,
        **kwargs,
    ):
        """
        Shorthand for log(message_type="error").

        See documentation of awsc.log() for parameter descriptions.
        """
        cls.log(
            message,
            summary,
            category,
            "error",
            subcategory=subcategory,
            resource=resource,
            set_message=set_message,
            **kwargs,
        )

    @classmethod
    def log_exception(
        cls,
        exception,
        category,
        subcategory=None,
        resource=None,
        set_message=True,
        **kwargs,
    ):
        """
        Shorthand for error(message=str(exception), summary=type(exception)). Also logs traceback.

        See documentation of awsc.log() for parameter descriptions.

        Parameters
        ----------
        exception : Exception
            Any exception object.
        """
        trace = "".join(traceback.format_tb(exception.__traceback__))
        cls.error(
            f"{str(exception)}\nTraceback:\n{trace}",
            type(exception).__name__,
            category,
            subcategory=subcategory,
            resource=resource,
            set_message=set_message,
            **kwargs,
        )

    @classmethod
    def clienterror(
        cls,
        error,
        summary,
        category,
        subcategory=None,
        resource=None,
        set_message=True,
        **kwargs,
    ):
        """
        Shorthand for extracting the AWS error message and passing it to error().

        See documentation of awsc.log() for parameter descriptions.

        Parameters
        ----------
        error : botocore.exceptions.ClientError
            Any boto3 ClientError derivative.
        """
        errtype = error.response["Error"]["Code"]
        errmsg = error.response["Error"]["Message"]
        cls.log(
            errmsg,
            f"{summary} | AWS: {errtype}",
            category,
            "error",
            subcategory=subcategory,
            resource=resource,
            set_message=set_message,
            **kwargs,
        )

    @classmethod
    def success(
        cls,
        message,
        summary,
        category,
        subcategory=None,
        resource=None,
        set_message=True,
        **kwargs,
    ):
        """
        Shorthand for log(message_type="success")

        See documentation of awsc.log() for parameter descriptions.
        """
        cls.log(
            message,
            summary,
            category,
            "success",
            subcategory=subcategory,
            resource=resource,
            set_message=set_message,
            **kwargs,
        )

    @classmethod
    def info(
        cls,
        message,
        summary,
        category,
        subcategory=None,
        resource=None,
        set_message=True,
        **kwargs,
    ):
        """
        Shorthand for log(message_type="info")

        See documentation of awsc.log() for parameter descriptions.
        """
        cls.log(
            message,
            summary,
            category,
            "info",
            subcategory=subcategory,
            resource=resource,
            set_message=set_message,
            **kwargs,
        )

    @classmethod
    def generic_api_call(
        cls,
        service,
        method,
        api_kwargs,
        summary,
        category,
        subcategory=None,
        success_template=None,
        resource=None,
        keys=None,
        **kwargs,
    ):
        """
        Executes an AWS API call. The event is logged and a response is returned.

        See documentation of awsc.log() for parameter descriptions.

        Parameters
        ----------
        service : str
            The boto3 provider for the API call.
        method : str
            The provider's method to call for the API call.
        api_kwargs : dict
            A map of keyword arguments to pass to the method call.
        success_template : str
            A format() string for the message to display when the API call is successful. No message is displayed if omitted.
            Formatting parameters:
            - The "resource" keyword and the 0 index for the value of the resource parameter.
            - All entries in api_kwargs.
            - All entries in kwargs.

        Returns
        -------
        dict
            A dict with two keys.
            - The key "Success" contains a boolean about whether the API call was a success.
            - The "Response" key is a dict. If the call was a success, it is the parsed JSON response from the AWS API. Otherwise, it contains
              the AWS error response. The error response contains the key "Error", which is a dict with the keys "Code" and "Message".
        """
        try:
            response = getattr(
                cls.Session.service_provider(service, keys=keys), method
            )(**api_kwargs)
            if success_template is not None:
                cls.success(
                    success_template.format(
                        resource, resource=resource, **api_kwargs, **kwargs
                    ),
                    summary,
                    category,
                    subcategory=subcategory,
                    resource=resource,
                    set_message=True,
                    api_provider=service,
                    api_method=method,
                    api_args=api_kwargs,
                    **kwargs,
                )
            return {"Success": True, "Response": response}
        except exceptions.ClientError as error:
            cls.clienterror(
                error,
                summary,
                category,
                subcategory=subcategory,
                resource=resource,
                set_message=True,
                api_provider=service,
                api_method=method,
                api_args=api_kwargs,
                **kwargs,
            )
            return {"Success": False, "Response": error.response}
        except Exception as error:
            cls.error(
                str(error),
                summary,
                category,
                subcategory=subcategory,
                resource=resource,
                set_message=True,
                api_provider=service,
                api_method=method,
                api_args=api_kwargs,
                **kwargs,
            )
            return {
                "Success": False,
                "Response": {
                    "Error": {
                        "Code": f"Python.{type(error).__name__}",
                        "Message": str(error),
                    }
                },
            }

    @classmethod
    def textfield_colors(cls, prefix):
        """
        Generates the commonly used color arguments for a textfield dialog given a prefix.

        Parameters
        ----------
        prefix : str
            The prefix for the colors.

        Returns
        -------
        dict
            A partial set of keyword arguments for DialogFieldText().
        """
        return {
            "color": cls.color(
                f"{prefix}_modal_dialog_textfield", "modal_dialog_textfield"
            ),
            "selected_color": cls.color(
                f"{prefix}_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
            "label_color": cls.color(
                f"{prefix}_modal_dialog_textfield_label", "modal_dialog_textfield_label"
            ),
            "label_min": 16,
        }

    @staticmethod
    def str_attrgetter(attr):
        """
        Returns a sort predicate like attrgetter that forces everything to be a string.

        Parameters
        ----------
        attr : str
            The attribute name for the attrgetter.

        Returns
        -------
        callable(dict) -> object
            Function which returns the named attribute of the parameter.
        """

        def fn(x):
            value = x[attr]
            if value is None:
                return ""
            return str(value)

        return fn

    @classmethod
    def default_border(cls, prefix, title, title_info=None):
        """
        Generates the default border object for the specified prefix, title and title_info.

        Parameters
        ----------
        prefix : str
            The color scheme prefix for the control which is requesting a border.
        title : str
            The title for the control.
        title_info : str
            The title information for the control.

        Returns
        -------
        awsc.termui.control.Border
            The generated default border.
        """
        return Border(
            cls.border("resource_list", "default"),
            cls.color(f"{prefix}_border", "generic_border"),
            title,
            cls.color(f"{prefix}_border_title", "border_title"),
            title_info,
            cls.color(f"{prefix}_border_title_info", "border_title_info"),
        )


class LogHolder:
    """
    Log stash object. Contains a list of log message from this and past runs. Interacts with the log storage yaml file to store and reload
    log messages.

    Does not need to be instantiated, Common holds an instance of this.

    Attributes
    ----------
    raw_entries : list(dict)
        A list of raw log entries. Raw log entries are in a dict-form, rather than as ListEntry objects.
    control : awsc.log.LogLister
        If the log list control is active, this is a reference to it. Just in case something loggable happens while looking at logs.
    """

    def __init__(self):
        self.raw_entries = []
        self.control = None
        self.parse_log_messages()

    def parse_log_messages(self):
        """
        Reads all raw entries from the disk storage and loads it into raw_entries.
        """
        file = Common.confdir() / "log.yaml"
        if file.exists():
            with file.open("r") as file:
                self.raw_entries = yaml.safe_load(file.read())
        else:
            self.raw_entries = []
        self.parse_raw_entries()

    def parse_raw_entries(self):
        """
        Parses raw entries into the LogLister control if one is active right now.
        """
        if self.control is not None:
            self.control.entries.clear()
            for entry in self.raw_entries:
                self.control.add_raw_entry(entry)
            self.control.sort()

    def write_raw_entries(self):
        """
        Writes the loaded raw entries to disk. Expunges log entries that go over log retention limits on lines or age.
        """
        file_path = Common.confdir() / "log.yaml"
        limit = Common.Configuration["log_retention"]["max_lines"]
        if limit > 0:
            max_idx = limit
        else:
            max_idx = len(self.raw_entries)
        age_limit = Common.Configuration["log_retention"]["max_age"]
        now = datetime.now(timezone.utc).timestamp()
        if age_limit > 0:
            for idx, raw_entry in enumerate(self.raw_entries):
                if now - raw_entry["timestamp"] > age_limit:
                    max_idx = idx
                    break

        self.raw_entries = self.raw_entries[:max_idx]

        with file_path.open("w") as file:
            file.write(yaml.dump(self.raw_entries))

    def attach(self, control):
        """
        Attaches the log stash to a LogLister control and populates it.

        Parameters
        ----------
        control : awsc.termui.LogLister
            The control to attach.
        """
        self.control = control
        for entry in self.raw_entries:
            self.control.add_raw_entry(entry)

    def detach(self):
        """
        Detaches a LogLister control from the log stash, probably because it was closed.
        """
        self.control = None

    def log(
        self,
        message,
        summary,
        category,
        message_type,
        subcategory=None,
        resource=None,
        set_message=True,
        **kwargs,
    ):
        """
        Adds a new log message to the log stash.

        See the documentation of Common.log() for the documentation of the parameters.
        """
        if set_message:
            color = "message_info"
            if message_type == "success":
                color = "message_success"
            elif message_type == "error":
                color = "message_error"
            Common.Session.set_message(message, Common.color(color))
        self.raw_entries.insert(
            0,
            {
                "summary": summary,
                "category": category,
                "subcategory": subcategory,
                "type": message_type,
                "message": message,
                "resource": resource,
                "timestamp": datetime.now(timezone.utc).timestamp(),
                "context": json.dumps(kwargs, default=datetime_hack),
            },
        )
        if self.control is not None:
            self.control.add_raw_entry(self.raw_entries[0])
        self.write_raw_entries()



"""
This module defines blocks that allow segmenting the terminal screen.
"""
from .common import Commons


class Block:
    """
    A Block object represents a section of the screen. A Block may contain additional Blocks or act as a control.

    Attributes
    ----------
    blocks : list
        A set of child Block objects.
    weight : int
        Sibling blocks are drawn in order from highest weight to lowest weight.
    parent : awsc.termui.block.Block
        The parent Block object. None indicates that the parent is the entire screen.
    alignment : awsc.termui.alignment.AbstractAnchor
        The anchor specifier for this Block within the parent Block.
    dimensions : awsc.termui.alignment.Dimension
        The size specifier for this Block.
    tag : str
        Tags allow grouping sibling Blocks, and removing them together.
    """

    def __init__(
        self, parent, alignment, dimensions, *args, weight=0, tag="default", **kwargs
    ):
        """
        Initializes a Block object.

        Parameters
        ----------
        parent : awsc.termui.block.Block
            The parent Block object. None indicates that the parent is the entire screen.
        alignment : awsc.termui.alignment.AbstractAnchor
            The anchor specifier for this Block within the parent Block.
        dimensions : awsc.termui.alignment.Dimension
            The size specifier for this Block.
        weight : int
            Sibling blocks are drawn in order from highest weight to lowest weight.
        tag : str
            Tags allow grouping sibling Blocks, and removing them together.
        """
        self.blocks = []
        self.weight = weight
        if parent is not None:
            parent.add_block(self)
        self.parent = parent
        self.alignment = alignment
        self.dimensions = dimensions
        self.tag = tag
        Commons.UIInstance.dirty = True
        self.on_become_frame_hooks = []

    def on_become_frame(self):
        for hook in self.on_become_frame_hooks:
            hook[0](**hook[1])

    def reparent(self):
        """
        Shorthand function for removing a Block from its parent and re-adding it.

        Useful for reordering sibling Blocks with equal weight, as reparenting will send the block atop all other sibling blocks with equal weight.

        Reparenting is also required if the weight of the block changes, as blocks are only sorted upon insertion.
        """
        self.parent.remove_block(self)
        self.parent.add_block(self)

    def topleft(self):
        """
        Returns the coordinates of the top left corner of the block.

        Returns
        -------
        tuple(int, int)
            The (x, y) coordinates of the top left corner of the block.
        """
        return self.alignment.topleft(self.dimensions, self.parent)

    def bottomright(self):
        """
        Returns the coordinates of the bottom right corner of the block.

        Returns
        -------
        tuple(int, int)
            The (x, y) coordinates of the bottom right corner of the block.
        """
        topleft = self.alignment.topleft(self.dimensions, self.parent)
        return (topleft[0] + self.dimensions[0], topleft[1] + self.dimensions[1])

    @property
    def width(self):
        """
        Read-only property for the width of the block.

        Returns
        -------
        int
            The width of the block in character cells.
        """
        corners = self.corners
        return corners[0][1] - corners[0][0] + 1

    @property
    def height(self):
        """
        Read-only property for the height of the block.

        Returns
        -------
        int
            The height of the block in character cells.
        """
        corners = self.corners
        return corners[1][1] - corners[1][0] + 1

    @property
    def w_in(self):
        """
        Read-only property for the inner width of the block. Subclasses may implement features which take away from the "usable" space
        of a block, therefore affecting its inner width without affecting its total width.

        Returns
        -------
        int
            The inner width of the block in character cells.
        """
        return self.width

    @property
    def corners(self):
        """
        Read-only property for calculating the four corners of the block.

        Returns
        -------
        tuple(tuple(int, int), tuple(int, int))
            Contains the horizontal bounds and vertical bounds of the block in the form of ((x0, x1), (y0, y1)) in character cell coordinates.
        """
        topleft = self.topleft()
        botright = (
            topleft[0] + self.dimensions[0] - 1,
            topleft[1] + self.dimensions[1] - 1,
        )
        return ((topleft[0], botright[0]), (topleft[1], botright[1]))

    @property
    def inner(self):
        """
        Read-only property for calculating the four inner corners of the block. Subclasses may implement features which take away from
        the "usable" space of a block, therefore affecting its inner corner coordinates without affecting its corners.

        Returns
        -------
        tuple(tuple(int, int), tuple(int, int))
            Contains the horizontal bounds and vertical inner corners of the block in the form of ((x0, x1), (y0, y1)) in character cell
            coordinates.
        """
        return self.corners

    def before_paint(self):
        """
        Hook function for acting on the block before any blocks are painted for the current frame.
        """
        for block in reversed(self.blocks):
            block.before_paint()

    def paint(self):
        """
        Hook function for painting the block. The block is expected to handle its own output through interacting with the UI object.
        """
        for block in reversed(self.blocks):
            block.paint()

    def add_block(self, block):
        """
        Adds a new child block to this block.

        When a new block is inserted, it is added as the topmost block for its own weight.

        Parameters
        ----------
        block : awsc.termui.block.Block
            The child block to insert.
        """
        Commons.UIInstance.dirty = True
        add_idx = 0
        for idx, existing_block in enumerate(self.blocks):
            if existing_block.weight > block.weight:
                add_idx = idx
                break
        block.parent = self
        self.blocks.insert(add_idx, block)

    def clear_blocks(self, tag=None):
        """
        Removes all blocks with the matching tag from the child blocks of this block.

        If no tag is provided, this function clears all child blocks.

        Parameters
        ----------
        tag : str
            The tag to search for when clearing blocks.
        """
        self.blocks = [
            block for block in self.blocks if tag is not None and block.tag != tag
        ]

    def remove_block(self, block):
        """
        Removes a specific block from this block.

        Parameters
        ----------
        block : awsc.termui.block.Block
            The block to remove from the children of this block.
        """
        try:
            self.blocks.remove(block)
        except ValueError:
            pass

    def input(self, key):
        """
        Input handler hook function. The function's return value indicates whether the processing of the input has been done
        by this or a child block and no further processing of the input should take place.

        The input will be piped to the lowest weight block first, proceeding in order to the highest weight block until a block
        reports that it handled the input successfully.

        Any block has the option of reacting to an input but not marking it as handled. This may be useful for key press debugging
        blocks.

        Not all keys have to be handled by the same block. A set of keys may be isolated as unhandled by any block other than a global
        hotkey handler block if needed.

        Parameters
        ----------
        key : blessed.keyboard.Keystroke
            The raw inkey() from the keyboard event.

        Returns
        -------
        bool
            True if the input was handled by this or any other child block, False otherwise.
        """
        for block in self.blocks:
            if block.input(key):
                return True
        return False
